/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FileInfoModel.h"

#include "CppUtil.h"
#include "FileIconProvider.h"
#include "FtpDefines.h"
#include "TimeStamp.h"
#include "XmlOptions.h"
#include "QOrderedSet.h"

#include <QIcon>
#include <QUrl>

//_______________________________________________________________
FileInfoModel::FileInfoModel( QObject* parent, BaseFileInfo::TypeFlags type ):
    BaseFileInfoModel<FileInfo>( parent ),
    Counter( "FileInfoModel" ),
    type_( type )
{ Debug::Throw( "FileInfoModel::FileInfoModel.\n" ); }

//__________________________________________________________________
Qt::ItemFlags FileInfoModel::flags(const QModelIndex &index) const
{

    Qt::ItemFlags flags( BaseFileInfoModel<FileInfo>::flags( index ) );
    if( !index.isValid() )
    {

        flags |= Qt::ItemIsDropEnabled;

    } else if( contains( index ) ) {

        // retrieve associated file info
        const FileInfo& fileInfo( get()[index.row()] );
        if( fileInfo.isNavigator() ) flags &= ~Qt::ItemIsSelectable;
        else {

            flags |= Qt::ItemIsDragEnabled;
            if( fileInfo.isFolder() ) flags |= Qt::ItemIsDropEnabled;

        }

    }

    return flags;

}

//__________________________________________________________________
QVariant FileInfoModel::data( const QModelIndex& index, int role ) const
{

    // check index, role and column
    if( !index.isValid() ) return QVariant();

    // retrieve associated file info
    const auto& fileInfo( get()[index.row()] );

    // return text associated to file and column
    switch( role )
    {

        case Qt::DisplayRole:
        {

            switch( index.column() )
            {

                case Permissions:
                return (fileInfo.isRemote() || fileInfo.type() & FileInfo::Navigator) ? QString():fileInfo.permissionsString();

                case Location:
                return QString( fileInfo.isLocal() ? tr( "Local" ): tr( "Remote" ) );

                case Status:
                return QString( fileInfo.isSuccess() ? tr( "Success" ): tr( "Failed" ) );

                case Action:
                return QString( fileInfo.isOverwrite() ? tr( "Overwrite" ): tr( "Skip" ) );

                default:
                return BaseFileInfoModel<FileInfo>::data( index, role );
            }
        }

        case Base::FileTypeRole:
        {
            int type( fileInfo.type() );
            if( clipboard_.mode() == FileSelectionClipboard::Cut && clipboard_.get().contains( fileInfo ) )
            { type |= BaseFileInfo::Clipped; }

            return type;
        }

        default:
        return BaseFileInfoModel<FileInfo>::data( index, role );

    }

}

//______________________________________________________________________
QStringList FileInfoModel::mimeTypes() const
{
    return _isRemote() ?
        Base::makeT<QStringList>({ BaseFileInfo::MimeType, "text/uri-list" }):
        Base::makeT<QStringList>({ BaseFileInfo::MimeType });

}

//______________________________________________________________________
QMimeData* FileInfoModel::mimeData(const QModelIndexList &indexes) const
{

    QOrderedSet<QString> filenames;
    QOrderedSet<FileInfo> fileInfoList;

    for( const auto& index:indexes )
    {

        if( index.isValid() && (flags(index)&Qt::ItemIsDragEnabled) )
        {
            FileInfo info( get(index) );

            fileInfoList.insert( info );
            filenames.insert( info.file() );
        }

    }

    if( filenames.empty() ) return 0;
    else {

        auto mimeData = new QMimeData;

        // fill text and url data
        QList<QUrl> urlList;
        QStringList fileList;
        for( const auto& filename:filenames )
        {
            const QString url( ( (type_ & FileInfo::Remote ) ? Ftp::Defines::remotePrefix : Ftp::Defines::localPrefix ) + "//" + filename );
            fileList.append( url );
            urlList.append( url );
        }

        mimeData->setText( fileList.join( "\n" ) );
        mimeData->setUrls( urlList );

        // fill Drag data. Use XML
        QDomDocument document;
        auto top = document.appendChild( document.createElement( Xml::FileInfoList ) ).toElement();
        for( const auto& fileInfo:fileInfoList )
        {

            if( fileInfo.file().isEmpty() ) continue;
            top.appendChild( fileInfo.domElement( document ) );

        }

        mimeData->setData( BaseFileInfo::MimeType, document.toByteArray() );
        return mimeData;

    }

}

//__________________________________________________________________
bool FileInfoModel::dropMimeData(const QMimeData* data , Qt::DropAction action, int, int, const QModelIndex& parent)
{

    Debug::Throw( "FileInfoModel::dropMimeData.\n" );

    // check action
    if( action == Qt::IgnoreAction) return true;

    // info list
    FileInfo::List fileInfoList;

    // check format
    if( data->hasFormat( BaseFileInfo::MimeType ) )
    {

        // get dropped file info (use XML)
        // dom document
        QDomDocument document;
        if( !document.setContent( data->data( BaseFileInfo::MimeType ), false ) ) return false;

        const auto docElement = document.documentElement();
        const auto node = docElement.firstChild();
        for( auto&& node = docElement.firstChild(); !node.isNull(); node = node.nextSibling() )
        {
            const auto element = node.toElement();
            if( element.isNull() ) continue;

            // special options
            if( element.tagName() == Xml::FileInfo )
            {

                const FileInfo fileInfo( element );
                if( fileInfo.file().isEmpty() || fileInfo.isNavigator() ) continue;
                fileInfoList.append( fileInfo );

            }
        }

    } else if( _isRemote() && data->hasUrls() ) {

        for( const auto& url:data->urls() )
        {

            #if QT_VERSION >= 0x040800
            // check that local file
            if( !url.isLocalFile() ) continue;
            #endif

            // get file and check existence
            const File file( url.path() );
            if( !file.exists() ) continue;

            FileInfo fileInfo( file );
            fileInfo.update();

            fileInfo.setLocal();
            if( file.isDirectory() ) fileInfo.setIsFolder();
            else fileInfo.setIsDocument();
            if( file.isLink() ) fileInfo.setIsLink();
            if( file.isBrokenLink() ) fileInfo.setIsBrokenLink();
            if( file.isHidden() ) fileInfo.setIsHidden();

            fileInfoList.append( fileInfo );

        }

    }

    if( fileInfoList.isEmpty() ) return false;

    // check destination
    FileInfo destination;
    if( parent.isValid() && get( parent ).isFolder() )
    {

        // get destination path, and add trailing "/" if missing
        destination = get( parent );
        destination.setFile( File( destination.file() ).addTrailingSlash() );

    } else destination.setType( type_ );

    // delayed transfer request emission
    // this fixes issues with nested event loops
    transferData_.set( destination, fileInfoList );
    QMetaObject::invokeMethod( this, "_requestTransfer", Qt::QueuedConnection );
    return true;

}

//__________________________________________________________________
QVariant FileInfoModel::headerData(int column, Qt::Orientation orientation, int role) const
{

    if( orientation == Qt::Horizontal && role == Qt::DisplayRole )
    {
        switch( column )
        {
            case Location: return tr( "Location" );
            case Status: return tr( "Status" );
            case Action: return tr( "Action" );
            default: break;
        }
    }

    // base class implementation
    return BaseFileInfoModel<FileInfo>::headerData( column, orientation, role );

}

//__________________________________________________________________
void FileInfoModel::setClipboard( FileSelectionClipboard clipboard )
{

    // remove files that are not of same local type
    if( clipboard.mode() == FileSelectionClipboard::Cut ) clipboard.removeLocal( _isLocal() );
    else clipboard.clear();

    // do nothing if no actual change
    if( clipboard_ == clipboard ) return;

    // if current is empty and new is either invalid or empty do nothing
    if( clipboard_.isEmpty() && clipboard.isEmpty() ) return;

    // update
    emit layoutAboutToBeChanged();
    clipboard_ = clipboard;
    emit layoutChanged();

}

//____________________________________________________________
void FileInfoModel::_sort( int column, Qt::SortOrder order )
{ std::sort( _get().begin(), _get().end(), SortFTor( (ColumnType) column, order ) ); }

//____________________________________________________________
QIcon FileInfoModel::_icon( const QModelIndex& index ) const
{

    if( !iconProvider_ ) return QIcon();
    else {
        const auto& fileInfo( get( index ) );
        return iconProvider_->icon( fileInfo, fileInfo.type() );
    }

}

//________________________________________________________
bool FileInfoModel::SortFTor::operator () ( const FileInfo& constFirst, const FileInfo& constSecond ) const
{

    // Navigator come first
    // Folders come next
    if( constFirst.isNavigator() ) return true;
    else if( constSecond.isNavigator() ) return false;
    else if(
        (constFirst.type()&(BaseFileInfo::Folder|BaseFileInfo::Document)) !=
        (constSecond.type()&(BaseFileInfo::Folder|BaseFileInfo::Document)) )
    { return constFirst.isFolder(); }

    FileInfo first( constFirst );
    FileInfo second( constSecond );
    if( order_ == Qt::DescendingOrder ) std::swap( first, second );

    switch( type_ )
    {

        case Location:
        {
            if( first.isLocal() && second.isRemote() ) return true;
            else if( second.isLocal() && first.isRemote() ) return false;
            else return first.file() < second.file();
        }

        case Status:
        return first.isSuccess() && !second.isSuccess();

        case Action:
        return first.isOverwrite() && !second.isOverwrite();

        default:
        return BaseFileInfoModel<FileInfo>::SortFTor::operator() ( constFirst, constSecond );

    }

}
