#ifndef FileInfoModel_h
#define FileInfoModel_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseFileInfoModel.h"
#include "Counter.h"
#include "FileInfo.h"
#include "FileSelectionClipboard.h"

#include <QMimeData>

class FileIconProvider;

//* FileInfo model. Stores file information for display in lists
class FileInfoModel : public BaseFileInfoModel<FileInfo>, private Base::Counter<FileInfoModel>
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    explicit FileInfoModel( QObject*, BaseFileInfo::TypeFlags = FileInfo::None );

    //* number of columns
    enum { nExtraColumns = 3 };

    //* column type enumeration
    enum ColumnType
    {
        Location = nColumns,
        Status,
        Action
    };

    //*@name accessors
    //@{
    // flags
    Qt::ItemFlags flags( const QModelIndex& ) const override;

    // return data for a given index
    QVariant data( const QModelIndex&, int ) const override;

    //* mime type
    QStringList mimeTypes() const override;

    //* mime data
    QMimeData* mimeData( const QModelIndexList& ) const override;

    //* header data
    QVariant headerData( int, Qt::Orientation, int = Qt::DisplayRole ) const override;

    //* number of columns for a given index
    int columnCount( const QModelIndex& = QModelIndex() ) const override
    { return nColumns + nExtraColumns; }

    //@}

    //*@name modifiers
    //@{

    //* set icon provider
    void setIconProvider( FileIconProvider* value )
    { iconProvider_ = value; }

    //* set clipboard
    void setClipboard( FileSelectionClipboard );

    //* drop mine data
    bool dropMimeData( const QMimeData*, Qt::DropAction, int, int, const QModelIndex& ) override;

    //@}

    //* used to sort FileInfo objects
    class SortFTor: public BaseFileInfoModel<FileInfo>::SortFTor
    {

        public:

        //* constructor
        explicit SortFTor( int type, Qt::SortOrder order ):
            BaseFileInfoModel<FileInfo>::SortFTor( type, order )
            {}

        //* prediction
        bool operator() ( const FileInfo&, const FileInfo& ) const override;

    };

    Q_SIGNALS:

    //* request transfer (from drag and drop)
    void transferRequested( const FileInfo&, const FileInfo::List& );

    protected:

    //* sort
    void _sort( int, Qt::SortOrder ) override;

    //* icon matching given model index
    QIcon _icon( const QModelIndex& ) const override;

    private Q_SLOTS:

    //* transfer request
    void _requestTransfer()
    { emit transferRequested( transferData_.destination(), transferData_.files() ); }

    private:

    //* true if local
    bool _isLocal() const
    { return !( type_ & FileInfo::Remote ); }

    //* true if remote
    bool _isRemote() const
    { return type_ & FileInfo::Remote; }

    //* base type (bitwise or of FileInfo::Type
    /** this is used to distinguish between local and remote list views */
    BaseFileInfo::TypeFlags type_ = 0;

    //* icon provider
    FileIconProvider* iconProvider_ = nullptr;

    //* clipboard
    FileSelectionClipboard clipboard_;

    //* data structure used for drag and drop
    class TransferData
    {
        public:

        //* constructor
        explicit TransferData()
        {}

        //* set
        void set( const FileInfo& destination, const FileInfo::List& files )
        {
            destination_ = destination;
            files_ = files;
        }

        //* destination
        const FileInfo& destination() const
        { return destination_; }

        //* files
        const FileInfo::List& files() const
        { return files_; }

        private:

        //* destination
        FileInfo destination_;

        //* files
        FileInfo::List files_;

    };

    //* transfer data
    TransferData transferData_;

};

#endif
