/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FileSelectionDialog.h"

#include "Debug.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QLabel>

//______________________________________________________________________
FileSelectionDialog::FileSelectionDialog( QWidget* parent, const QString& label ):
    QuestionDialog( parent, label ),
    model_( this, FileInfo::None )
{

    Debug::Throw( "FileSelectionDialog::FileSelectionDialog.\n" );
    setSizeGripEnabled( true );
    setOptionName( "FILE_SELECTION_DIALOG" );
    mainWidget_ = new QWidget( this );

    // main (working) layout
    mainLayout_ = new QVBoxLayout;
    mainLayout_->setSpacing( 10 );
    mainLayout_->setMargin( 0 );
    mainWidget_->setLayout( mainLayout_ );

    // horizontal layout for list and comments
    auto hLayout = new QHBoxLayout;
    hLayout->setSpacing( 5 );
    hLayout->setMargin( 0 );
    mainLayout_->addLayout( hLayout, 1 );

    // create list
    hLayout->addWidget( list_ = new TreeView( mainWidget_ ) );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->setModel( &model_ );
    list_->setMask( XmlOptions::get().get<int>( "LOCAL_LIST_MASK" ) );

    // work layout
    auto vLayout = new QVBoxLayout;
    vLayout->setSpacing(0);
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout );

    workLayout_ = new QVBoxLayout;
    workLayout_->setSpacing( 5 );
    workLayout_->setMargin( 0 );
    vLayout->addLayout( workLayout_ );
    vLayout->addStretch(1);

    // select all
    QPushButton* button;
    workLayout_->addWidget( button = new QPushButton( tr( "Select All" ), mainWidget_ ) );
    button->setToolTip( tr( "Select all files in list" ) );
    connect( button, SIGNAL(clicked()), list_, SLOT(selectAll()) );

    // deselect all
    workLayout_->addWidget( clearSelectionButton_ = new QPushButton( tr( "Clear Selection" ), mainWidget_ ) );
    clearSelectionButton_->setToolTip( tr( "Deselect all files in list" ) );
    connect( clearSelectionButton_, SIGNAL(clicked()), list_, SLOT(clearSelection()) );

    // details
    setDetails( mainWidget_ );
    showDetails();

    // list connection
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    _updateButtons();

}

//____________________________________________________________________
void FileSelectionDialog::setFiles( const File& localPath, const File& remotePath, const FileInfo::List& files )
{

    Debug::Throw( "FileSelectionDialog::setFiles.\n" );

    // store path
    localPath_ = localPath;
    remotePath_ = remotePath;

    // add all conflicting items to list
    FileInfo::List local;
    std::transform( files.begin(), files.end(), std::back_inserter( local ),
        [this]( const FileInfo& fileInfo )
        {
            FileInfo copy( fileInfo );
            copy.setFile( File( fileInfo.file() ).addPath( fileInfo.isLocal() ? localPath_:remotePath_ ) );
            return copy;
        });

    model_.set( local );
    list_->resizeColumnToContents( FileInfoModel::Action );
    list_->resizeColumnToContents( FileInfoModel::FileName );
    list_->selectAll();

}

//____________________________________________________________________
void FileSelectionDialog::updateFiles( FileInfo::List& files )
{

    Debug::Throw( "FileSelectionDialog::updateFiles.\n" );

    // retrieve all items and put corresponding file info into output list
    auto local( model_.get() );

    // should loop over files in input list
    for( auto& fileInfo:files )
    {

        // look for matching FileInfo in input list
        fileInfo.setFile( File( fileInfo.file() ).addPath( fileInfo.isLocal() ? localPath_:remotePath_ ) );

        // this will not work because the path has been modified
        auto fileIter = std::find_if( local.begin(), local.end(), FileInfo::SameFileFTor( fileInfo.file() ) );
        if( fileIter == local.end() ) continue;

        // update action
        fileInfo.setStatus( FileInfo::Overwrite, fileIter->isOverwrite() );

    }

    // some debugging output
    if( Debug::level() >= 1 )
    {
        Debug::Throw() << endl;
        for( const auto& fileInfo:files )
        {
            if( !fileInfo.isConflict() ) Debug::Throw() << "FileSelectionDialog::updateFiles - file: " << fileInfo.file() << " no conflict" << endl;
            else Debug::Throw() <<  "FileSelectionDialog::updateFiles - file: " << fileInfo.file() << (fileInfo.isOverwrite() ? " overwrite":" skip" ) << endl;
        }
    }

    return;

}

//____________________________________________________________________
void FileSelectionDialog::_updateButtons()
{
    Debug::Throw( "FileSelectionDialog::_updateButtons.\n" );
    bool hasSelection( !list_->selectionModel()->selectedRows().empty() );
    clearSelectionButton_->setEnabled( hasSelection );
}
