#ifndef FileSelectionFrame_h
#define FileSelectionFrame_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CustomComboBox.h"
#include "FileInfoModel.h"
#include "FileSelectionClipboard.h"
#include "FtpMacro.h"
#include "TextSelection.h"
#include "XmlOptions.h"

#include <QAbstractItemView>
#include <QItemSelectionModel>
#include <QMenu>

#include <memory>

// forward declarations
class PathEditor;
class BaseFileInfoToolTipWidget;

//* document/folder navigation
class FileSelectionFrame: public QWidget, private Base::Counter<FileSelectionFrame>
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    explicit FileSelectionFrame( QWidget*, BaseFileInfo::TypeFlags );

    //*@name accessors
    //@{

    //* current path
    FileInfo path() const;

    //* type
    BaseFileInfo::TypeFlags type() const
    { return type_; }

    //* selected files
    FileInfo::List selection() const;

    //* text selection
    TextSelection textSelection() const;

    //* returns true if this widget owns argument
    bool owns( QWidget* ) const;

    //* active state
    bool isActive() const
    { return active_; }

    //* list
    QAbstractItemView& list() const
    { return *list_; }

    //* model
    const FileInfoModel& model() const
    { return model_; }

    //* view mode
    enum class ViewMode
    {
        Details,
        Icons
    };

    //* view mode
    ViewMode viewMode() const
    { return viewMode_; }

    //* path editor
    PathEditor& pathEditor() const
    { return *pathEditor_; }

    //* contents description
    QString description() const;

    //@}

    //*@name modifiers
    //@{

    //* home path
    void setHomePath( const File& ) const;

    //* root path list
    void setRootPathList( const File::List& ) const;

    //* current path
    void setPath( const File& );

    // clear
    void clear();

    //* flags
    enum Flag
    {
        None = 0,
        Visible = 1<<1,
        Selected = 1<<2
    };

    Q_DECLARE_FLAGS( Flags, Flag );

    //* add File in list
    void add( const FileInfo&, Flags = None );

    //* add Files in list
    void add( const FileInfo::List& files, Flags flags = None )
    { for( const auto& file:files ) add( file, flags ); }

    //* remove File from list
    void remove( const FileInfo& );

    //* remove Files from list
    void remove( const FileInfo::List& files )
    { for( const auto& file:files ) remove( file ); }

    //* active state
    bool setActive( bool value )
    {
        if( active_ == value ) return false;
        active_ = value;
        _updateBackgroundColors();
        return true;
    }

    //* mode
    bool setViewMode( ViewMode );

    //* model
    FileInfoModel& model()
    { return model_; }

    //* find
    void find( TextSelection );

    //@}

    //*@name actions
    //@{

    //* toggle visibility
    QAction& visibilityAction() const
    { return *visibilityAction_; }

    //* previous directory
    QAction& reloadAction() const
    { return *reloadAction_; }

    //* previous directory
    QAction& previousDirectoryAction() const
    { return *previousDirectoryAction_; }

    //* next directory
    QAction& nextDirectoryAction() const
    { return *nextDirectoryAction_; }

    //* directory above
    QAction& parentDirectoryAction() const
    { return *parentDirectoryAction_; }

    //* home directory
    QAction& homeDirectoryAction() const
    { return *homeDirectoryAction_; }

    // new folder
    QAction& newFolderAction() const
    { return *newFolderAction_; }

    //* cut selection
    QAction& cutAction() const
    { return *cutAction_; }

    //* copy selection
    QAction& copyAction() const
    { return *copyAction_; }

    //* paste clipboard
    QAction& pasteAction() const
    { return *pasteAction_; }

    //* paste clipboard
    QAction& pasteInFolderAction() const
    { return *pasteInFolderAction_; }

    //* rename selection
    QAction& renameAction() const
    { return *renameAction_; }

    //* remove selection
    QAction& removeAction() const
    { return *removeAction_; }

    //* open
    QAction& openAction() const
    { return *openAction_; }

    //* file properties
    QAction& filePropertiesAction() const
    { return *filePropertiesAction_; }

    //@}

    Q_SIGNALS:

    //* selection chaged
    void selectionChanged();

    //* request for opening a file
    void fileSelected( FileInfo );

    //* request for opening multiple files
    void openFilesRequested( const FileInfo::List& );

    //* clipboard changed
    void clipboardChanged( const FileSelectionClipboard& );

    //* emitted when actions enable state has changed
    void actionsChanged();

    //* request command
    void macroRequested( const Ftp::Macro& );

    //* request commands
    void macrosRequested( const Ftp::Macro::List& );

    //* request paste from clipboard
    void pasteClipboardRequested( const FileInfo& );

    //* request file information
    void requestFileInformation( const FileInfo&, const QIcon& );

    public Q_SLOTS:

    //* hidden files
    void setShowHiddenFiles( bool value )
    {
        if( showHiddenFiles_ == value ) return;
        showHiddenFiles_ = value;
        _reload();
    }

    //* navigator
    void setShowNavigator( bool value )
    {
        const FileInfo navigator( File(".."), BaseFileInfo::TypeFlags( FileInfo::Folder|FileInfo::Navigator|type_ ) );
        if( value ) model_.add( navigator );
        else model_.remove( navigator );
    }

    //* clear selection
    void clearSelection();

    //* select item
    void select( const FileInfo& );

    private Q_SLOTS:

    //* index activated
    void _indexActivated( const QModelIndex& );

    //* reload current directory
    void _reload();

    //* new folder
    void _newFolder();

    //* select home directory
    void _homeDirectory();

    //* cut
    void _cut();

    //* copy
    void _copy();

    //* paste
    void _paste();

    //* paste
    void _pasteInFolder();

    //* rename
    void _rename();

    //* remove
    void _remove();

    //* open
    void _open();

    //* file properties
    void _fileProperties();

    //* update list menu
    void _updateContextMenu( const QPoint& );

    //* update actions, based on selection
    void _updateActions();

    //* navigation actions
    void _updateNavigationActions();

    //* show tooltip
    void _showToolTip( const QModelIndex& );

    //* toggle visibility
    void _toggleVisibility( bool value )
    {
        setVisible( value );
        XmlOptions::get().set<bool>( visibilityOptionName_, value );
    }

    private:

    //* true if local
    bool _isLocal() const
    { return !( type_ & FileInfo::Remote ); }

    //* true if remote
    bool _isRemote() const
    { return type_ & FileInfo::Remote; }

    //* setup list
    void _setupList();

    //* show hidden files
    bool _showHiddenFiles() const
    { return showHiddenFiles_; }

    //* update background colors
    void _updateBackgroundColors() const;

    //* actions
    void _installActions();

    //* path line editor
    PathEditor* pathEditor_ = nullptr;

    //* tooltip manager
    BaseFileInfoToolTipWidget* toolTipWidget_ = nullptr;

    //* model
    FileInfoModel model_;

    //* selection model
    QItemSelectionModel selectionModel_;

    //* list of files
    std::unique_ptr<QAbstractItemView> list_;

    //*@name menus
    //@{

    //* column selection
    QMenu* columnSelectionMenu_ = nullptr;

    //* column sorting
    QMenu* columnSortingMenu_ = nullptr;

    //@}

    //*@name actions
    //@{

    //* toggle visibility
    QAction* visibilityAction_ = nullptr;

    //* reload
    QAction* reloadAction_ = nullptr;

    //* previous directory action
    QAction* previousDirectoryAction_ = nullptr;

    //* next directory action
    QAction* nextDirectoryAction_ = nullptr;

    //* parent directory action
    QAction* parentDirectoryAction_ = nullptr;

    //* home directory action
    QAction* homeDirectoryAction_ = nullptr;

    //* new folder action
    QAction* newFolderAction_ = nullptr;

    //* cut selection action
    QAction* cutAction_ = nullptr;

    //* copy selection action
    QAction* copyAction_ = nullptr;

    //* paste
    QAction* pasteAction_ = nullptr;

    //* paste in folder action
    QAction* pasteInFolderAction_ = nullptr;

    //* rename action
    QAction* renameAction_ = nullptr;

    //* remove selection
    QAction* removeAction_ = nullptr;

    //* open selection
    QAction* openAction_ = nullptr;

    //* file properties
    QAction* filePropertiesAction_ = nullptr;

    //@}

    //* base type (bitwise or of FileInfo::Type
    /*! this is used to distinguish between local and remote list views */
    BaseFileInfo::TypeFlags type_;

    //* visibility option name
    QString visibilityOptionName_;

    //* view mode option name
    QString viewModeOptionName_;

    //* mode
    ViewMode viewMode_ = ViewMode::Icons;

    //* show/hide hidden files
    bool showHiddenFiles_ = false;

    //* active state
    bool active_ = false;

};

Q_DECLARE_OPERATORS_FOR_FLAGS( FileSelectionFrame::Flags )

#endif
