#ifndef FtpCommand_h
#define FtpCommand_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "CppUtil.h"
#include "FileInfo.h"
#include "Functors.h"

#include <QFtp>
#include <QList>
#include <QHash>

namespace Ftp
{
    class Command final: private Base::Counter<Command>
    {

        public:

        //* enumeration for command type
        enum class Type
        {
            None,
            ConnectToHost,
            Login,
            Close,
            ChangeDirectory,
            CreateDirectory,
            ShowDirectory,
            Copy,
            Rename,
            Remove
        };

        //* list of commands
        using List = QList<Command>;

        #if QT_VERSION < 0x040800
        //* constructor
        explicit Command( Type, std::initializer_list<FileInfo>&& );
        #endif

        //* constructor
        explicit Command( Type = Type::None, FileInfo::List&& arguments = FileInfo::List() );

        //*@name accessors
        //@{

        //* type
        Type type() const
        { return type_; }

        //* return command name
        QString name() const
        { return _names()[type_]; }

        //* id
        int id() const
        { return id_; }

        //* ftp id
        int ftpId() const
        { return ftpId_; }

        //* argument size
        int argumentCount() const
        { return arguments_.size(); }

        //* get argument
        const FileInfo& argument( int index ) const
        { return arguments_[index]; }

        //@}

        //*@name modifiers
        //@{

        //* id
        void setFtpId( int ftpId )
        { ftpId_ = ftpId; }

        //* add argument
        Command& addArgument( const FileInfo& argument )
        {
            arguments_.append(argument);
            return *this;
        }

        //* clear arguments
        void clearArguments()
        { arguments_.clear(); }

        //@}

        //* retrieve command with same type
        using SameTypeFTor = Base::Functor::Unary<Command, Type, &Command::type>;

        private:

        //* static id counter
        static int& _counter();

        //* command names
        using NameMap = QHash< Type, QString >;
        static const NameMap& _names();

        //* command id
        int id_ = 0;

        //* ftp id
        int ftpId_ = 0;

        //* command type
        Type type_ = Type::None;

        //* arguments
        FileInfo::List arguments_;

    };

    using ::qHash;

    //* equal to operator
    inline bool operator == (const Command& first, const Command& second)
    { return first.id() == second.id(); }

    //* less than to operator
    inline bool operator < (const Command& first, const Command& second)
    { return first.id() < second.id(); }

}

#endif
