/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FtpInterface.h"

#include "Command.h"
#include "ErrorDialog.h"
#include "QtUtil.h"
#include "TextFormat.h"
#include "Util.h"
#include "QOrderedSet.h"

#include <QApplication>
#include <QDesktopServices>
#include <QUrl>

#include <algorithm>
#include <numeric>

namespace Ftp
{
    //_______________________________________________________________________
    Interface::Interface( QObject* parent ):
        QObject( parent ),
        Counter( "Ftp::Interface" ),
        ftp_( new QFtp( this ) ),
        thread_( new FileThread( this ) )
    {

        Debug::Throw( "Ftp::Interface::Interface.\n" );
        connect( ftp_, SIGNAL(commandStarted(int)), SLOT(_commandStarted(int)) );
        connect( ftp_, SIGNAL(commandFinished(int,bool)), SLOT(_commandCompleted(int,bool)) );
        connect( ftp_, SIGNAL(dataTransferProgress(qint64,qint64)), SLOT(_progress(qint64,qint64)) );
        connect( ftp_, SIGNAL(listInfo(QUrlInfo)), SLOT(_addRemoteFile(QUrlInfo)) );

        // thread connections
        connect( thread_, SIGNAL(filesAvailable(const File::List&)), SLOT(_processFiles(const File::List&)) );
        connect( thread_, SIGNAL(sizeAvailable(qint64)), SLOT(_processSize(qint64)) );
        connect( thread_, SIGNAL(finished()), SLOT(_commandCompleted()) );

    }

    //_________________________________________________________________
    void Interface::addMacro( const Macro& macro )
    {

        Debug::Throw() << "Ftp::Interface::addMacro - macro: " << macro.name() << endl;

        // check if empty
        const bool empty( macros_.empty() );

        // add macro
        macros_.append( macro );

        QString buffer;
        emit macroAdded( macro );

        if( empty )
        {
            emit busyStateChanged( true );
            QMetaObject::invokeMethod( this, "busy", Qt::QueuedConnection );
            QMetaObject::invokeMethod( this, "_runMacro", Qt::QueuedConnection );
        }

    }

    //_________________________________________________________________
    void Interface::addMacros( const Macro::List& macros )
    {


        // check if empty
        const bool empty( macros_.empty() );

        // add macro
        for( const auto& macro:macros )
        {
            macros_.append( macro );
            emit macroAdded( macro );
        }

        // run first
        if( empty )
        {
            emit busyStateChanged( true );
            QMetaObject::invokeMethod( this, "busy", Qt::QueuedConnection );
            QMetaObject::invokeMethod( this, "_runMacro", Qt::QueuedConnection );
        }

    }

    //_________________________________________________________________
    void Interface::setShowHiddenFiles( bool value )
    {
        Debug::Throw() << "Ftp::Interface::setShowHiddenFiles - value: " << value << endl;
        showHiddenFiles_ = value;
    }

    //_________________________________________________________________
    void Interface::_runMacro()
    {

        // do nothing if no macros are in the pile
        if( macros_.empty() )
        {
            QMetaObject::invokeMethod( this, "ready", Qt::QueuedConnection );
            emit busyStateChanged( false );
            return;
        }

        // store macro and path
        Macro& macro( macros_.front() );

        Debug::Throw() << "Ftp::Interface::_runMacro - macro: " << macro.name() << endl;

        emit macroStarted( macro );

        // open connection
        switch( macro.type() )
        {

            // add
            case Macro::Type::Add:
            {

                for( const auto& fileInfo:macro.arguments() )
                {

                    _addFileRecursive( fileInfo );
                    const QString buffer = tr( "Adding '%1' to cache\n" ).arg( fileInfo.file() );
                    emit logAvailable( buffer );

                }

                _macroCompleted();
                break;
            }

            // clear
            case Macro::Type::Clear:
            {

                const QString buffer( tr( "Clearing cache\n" ) );
                emit logAvailable( buffer, Format::Bold );

                _clearFilesRecursive();
                _clearFiles();
                _macroCompleted();
                break;
            }

            // connection
            case Macro::Type::Connect:
            {

                // retrieve arguments
                const FileInfo host( macro.argument(0) );
                const FileInfo port( macro.argument(1) );
                const FileInfo user( macro.argument(2) );
                const FileInfo password( macro.argument(3) );

                const QString buffer = tr( "Connecting to host '%1'\n" ).arg( host.file() );
                emit logAvailable( buffer, Format::Bold );

                if( ftp_->state() != QFtp::Unconnected )
                { _addCommand( Command( Command::Type::Close ) ); }
                _addCommand( Command( Command::Type::ConnectToHost, { host, port } ) );
                _addCommand( Command( Command::Type::Login, { user, password } ) );
                break;
            }

            // close connection
            case Macro::Type::CloseConnection:
            {

                const QString buffer( tr( "Closing connection\n" ) );
                emit logAvailable( buffer, Format::Bold );

                _addCommand( Command( Command::Type::Close ) );
                break;
            }

            // abort
            case Macro::Type::Abort:
            {

                const QString buffer( tr( "Aborting ftp commands\n" ) );
                emit logAvailable( buffer, Format::Bold );

                ftp_->abort();
                _macroCompleted();
                break;
            }

            // show directory macro
            case Macro::Type::ShowDirectory:
            {

                const FileInfo path( macro.argument(0) );
                const QString buffer = tr( "Fetching content of directory '%1'\n" ).arg( path.file() );
                emit logAvailable( buffer, Format::Bold );

                if( path.isRemote() )  _addCommand( Command( Command::Type::ChangeDirectory, { path } ) );
                _addCommand( Command( Command::Type::ShowDirectory, { path } ) );

                break;
            }

            // create directory
            case Macro::Type::CreateDirectory:
            {


                // retrieve local and remote path
                const FileInfo newPathInfo( macro.argument(1) );
                const QString buffer = tr( "Creating directory '%1'\n" ).arg( newPathInfo.file() );
                emit logAvailable( buffer, Format::Bold );

                if( newPathInfo.isRemote() )
                {

                    const File basePath( macro.argument(0).file() );
                    File newPath( newPathInfo.file() );
                    _removePath( newPath, basePath );

                    // need to create all subdirectories that match the path
                    for( int position = 1; ( position = newPath.get().indexOf( "/", position ) ) >= 0; ++position )
                    {
                        File subPath( newPath.get().left( position ) );
                        _addCommand( Command( Command::Type::CreateDirectory, { FileInfo( subPath.addPath( basePath ), FileInfo::Folder|FileInfo::Remote ) } ) );
                    }

                }

                _addCommand( Command( Command::Type::CreateDirectory, { newPathInfo } ) );
                break;

            }


            // create directory
            case Macro::Type::Rename:
            {

                // retrieve arguments
                const FileInfo& source( macro.argument(0) );
                const FileInfo& destination( macro.argument(1) );
                const QString buffer = tr( "Renaming file '%1'\n" ).arg( source.file() );
                emit logAvailable( buffer, Format::Bold );

                _addCommand( Command( Command::Type::Rename, { source, destination } ) );

                break;

            }

            // transfer
            case Macro::Type::CopyRecursive:
            {

                const QString buffer( tr( "Copying cached files\n" ) );
                emit logAvailable( buffer, Format::Bold );

                // compute total size to transfer
                _computeTotalSize();
                emit maxProgressAvailable( total_ );
                emit progressAvailable( 0 );

                // create list of commands
                Command::List commands;

                const FileInfo sourcePath( macro.argument(0) );
                const FileInfo destinationPath( macro.argument(1) );

                // loop over recursive files
                for( auto source:filesRecursive_ )
                {

                    // skip navigator and files that must be skipped
                    if( source.isNavigator() || ( source.isConflict() && !source.isOverwrite() ) ) continue;

                    // generate destination file
                    FileInfo destination( source );
                    File newPath( source.file() );
                    _removePath( newPath, sourcePath.file() );
                    destination.setFile( newPath.addPath( destinationPath.file() ) );

                    if( destinationPath.isLocal() ) destination.setLocal();
                    else destination.setRemote();

                    // when copying link to remote destination, use real file
                    if( source.isLocal() && source.isLink() && destination.isRemote() )
                    {
                        if( source.isBrokenLink() )
                        {

                            const QString buffer = tr( "File '%1' is a broken link. <Copy> aborted.\n" ).arg( source.file() );
                            emit logAvailable( buffer );
                            continue;

                        } else {

                            // use real file, update source accordingly
                            source.setFile( source.file().readLink() );
                            source.setIsLink( false );
                            source.update();

                            destination.setIsLink( false );
                            destination.updateFrom( source );

                        }
                    }

                    // add proper command
                    // links are copied unchanged, when destination is local
                    if( destination.isDocument() || (destination.isLink() && destination.isLocal() ) ) commands.append( Command( Command::Type::Copy, { source, destination } ) );
                    else commands.append( Command( Command::Type::CreateDirectory, { destination } ) );

                }

                // submit commands
                if( !commands.isEmpty() ) _addCommands( commands );
                else _macroCompleted();

            }

            break;

            // transfer
            case Macro::Type::Remove:
            {

                const QString buffer( tr( "Removing cached files\n" ) );
                emit logAvailable( buffer, Format::Bold );

                // compute total size to remove
                _computeTotalSize();
                emit maxProgressAvailable( total_ );
                emit progressAvailable( 0 );

                // loop through files recursive and properly propagates failure flags
                FileInfo::List failed;
                for( const auto& fileInfo:filesRecursive_ )
                { if( fileInfo.isFailed() ) failed.append( fileInfo ); }

                for( auto& fileInfo:filesRecursive_ )
                {
                    if( fileInfo.isFailed() ) continue;

                    // see if current is a base directory of a failed file
                    // if so, tag as failed as well
                    for( const auto& failedFileInfo:failed )
                    { if( failedFileInfo.file().startsWith( fileInfo.file() ) ) fileInfo.setFailed(); }
                }

                // retrieve local and remote path
                const FileInfo path( macro.argument(0) );

                // list of commands
                Command::List commands;

                // loop backwards through filesRecursive and emit remove command for non failed jobs
                FileInfo::ListIterator iter( filesRecursive_ );
                iter.toBack();
                while( iter.hasPrevious() )
                {

                    const FileInfo& fileInfo( iter.previous() );
                    if( !fileInfo.isFailed() && fileInfo.location() == path.location() )
                    { commands.append( Command( Command::Type::Remove, { fileInfo } ) ); }

                }

                // submit commands
                if( !commands.isEmpty() ) _addCommands( commands );
                else _macroCompleted();

                break;

            }

            // transfer
            case Macro::Type::Open:
            {

                const QString buffer( tr( "Opening cached files\n" ) );
                emit logAvailable( buffer, Format::Bold );

                // compute total size to remove
                _computeTotalSize();
                emit maxProgressAvailable( total_ );
                emit progressAvailable( 0 );

                // list over recursive files (backward)
                bool hasRemote( false );

                // retrieve local and remote path
                const File localPath( macro.argument(0).file() );
                const File remotePath( macro.argument(1).file() );

                FileInfo::ListIterator iter( filesRecursive_ );
                iter.toBack();
                while( iter.hasPrevious() )
                {
                    const FileInfo& fileInfo( iter.previous() );
                    File file( fileInfo.file() );
                    if( fileInfo.isNavigator() ) continue;

                    if( fileInfo.isRemote() )
                    {

                        hasRemote = true;
                        Q_ASSERT( fileInfo.isDocument() );

                        // create matching local path name
                        _removePath( file, remotePath );

                        // temporary path
                        emit scratchFileCreated( File( Util::tmp() + "/qftp" ) );
                        emit scratchFileCreated( File( Util::tmp() + "/qftp/" + Util::user() ) );

                        // create local filename
                        File tmpFile( Util::tmp() + "/qftp/" + Util::user() + file.addPath( remotePath, false ) );

                        // create associated directory if not existent
                        if( !tmpFile.path().exists() )
                        {

                            Command command = Command( Command::Type::CreateDirectory, { FileInfo( tmpFile.path(), FileInfo::Folder ) } );
                            emit commandAdded( command );

                            File tmpPath( "qftp/" + Util::user() + file.addPath( remotePath, false ).path( false ) );

                            const bool error( !File( Util::tmp() ).createDirectory( tmpPath ) );
                            emit scratchFileCreated( tmpPath );

                            emit commandCompleted( command, error );

                        }

                        // process transfer
                        _addCommand( Command( Command::Type::Copy, { fileInfo, FileInfo( tmpFile, FileInfo::Document ) } ) );


                    } else if( fileInfo.isLocal() ) {

                        // run command on local file
                        File file( fileInfo.file() );

                        if( macro.hasArgument(2) )
                        {

                            // use specified command if any
                            File command( macro.argument(2).file() );
                            Base::Command( { command, fileInfo.file() } ).run( localPath );

                        } else {

                            // use default application otherwise
                            QDesktopServices::openUrl( QUrl::fromEncoded( QString( "file://%1" ).arg( fileInfo.file() ).toLatin1() ) );

                        }

                    }

                }

                // if no remote files found, the macro is completed
                if( !hasRemote ) _macroCompleted();
            }
            break;

            case Macro::Type::SizeRecursive:
            case Macro::Type::ListRecursive:
            {

                if( macro.type() == Macro::Type::SizeRecursive ) _resetTotalSize();

                // append current directory
                const FileInfo path( macro.argument(0) );
                _addFileRecursive( path );

                if( path.isFolder() )
                {

                    const QString buffer = tr( "Fetching content of directory '%1'\n" ).arg( path.file() );
                    emit logAvailable( buffer, Format::Bold );

                } else {

                    const QString buffer = tr( "Adding file '%1' to cache\n" ).arg( path.file() );
                    emit logAvailable( buffer, Format::Bold );

                }

                if( !path.isFolder() )
                {

                    _macroCompleted();

                } else {

                    if( path.isRemote() ) _addCommand( Command( Command::Type::ChangeDirectory, { path } ) );
                    _addCommand( Command( Command::Type::ShowDirectory, { path } ) );

                }

            }
            break;

            // check conflicts
            case Macro::Type::CheckConflicts:
            {

                const QString buffer( tr( "Checking conflicts\n" ) );
                emit logAvailable( buffer, Format::Bold );

                // retrieve local and remote path
                const FileInfo& sourcePathInfo( macro.argument(0) );
                const FileInfo& destinationPathInfo( macro.argument(1) );

                // keep track of path to list for conflicts
                QOrderedSet<File> remotePathSet;

                // loop over recursive files
                for( auto& fileInfo:filesRecursive_ )
                {

                    auto file( fileInfo.file() );
                    auto copy = fileInfo;
                    _removePath( file, sourcePathInfo.file() );
                    copy.setFile( file.addPath( destinationPathInfo.file() ) );
                    if( destinationPathInfo.isRemote() ) copy.setRemote();
                    else copy.setLocal();

                    if( copy.isLocal() )
                    {
                        Debug::Throw()
                            << "Ftp::Interface::_runMacro - checkConflicts"
                            << " source: " << fileInfo
                            << " destination: " << fileInfo.file() << endl;

                        if( copy.file().exists() )
                        {
                            // by default, skip file if it is a directory or if it is older
                            fileInfo.setStatus( FileInfo::Conflict, true );
                            if( fileInfo.isFolder() || fileInfo.isOlder( copy.file().lastModified() ) ) fileInfo.setStatus( FileInfo::Overwrite, false );
                        }

                    } else {

                        // get matching remote path and store
                        remotePathSet.insert( copy.file().path( false ) );
                        Debug::Throw()
                            << "Ftp::Interface::_runMacro - checkConflicts"
                            << " source: " << fileInfo
                            << " destination: " << copy.file() << endl;

                    }

                }

                // if no local files are found (to check against remote), the macro is finished
                if( remotePathSet.isEmpty() ) _macroCompleted();
                else {

                    for( const auto& file:remotePathSet )
                    { _addCommand( Command( Command::Type::ShowDirectory, { FileInfo( file, FileInfo::Folder | FileInfo::Remote ) } ) ); }

                }
            }
            break;

            // checkFiles macro
            case Macro::Type::CheckFilesCopy:
            case Macro::Type::CheckFilesMoved:
            {

                // retrieve local and remote path
                const FileInfo& sourcePathInfo( macro.argument(0) );
                const FileInfo& destinationPathInfo( macro.argument(1) );
                const QString buffer( tr( "Checking transfers\n" ) );
                emit logAvailable( buffer, Format::Bold );

                // keep track of path to list for conflicts
                QOrderedSet<File> remotePathSet;

                // list of updated files
                FileInfo::List updatedFiles;

                // loop over recursive files
                for( auto& fileInfo:filesRecursive_ )
                {

                    // if action is "skip" set success to true
                    if( !fileInfo.isOverwrite() )
                    {
                        fileInfo.setSuccess();
                        continue;
                    }

                    // retrieve associated file
                    // check location
                    auto copy( fileInfo );
                    auto file( fileInfo.file() );
                    _removePath( file, sourcePathInfo.file() );
                    copy.setFile( file.addPath( destinationPathInfo.file() ) );
                    if( destinationPathInfo.isRemote() ) copy.setRemote();
                    else copy.setLocal();

                    if( copy.isLocal() )
                    {

                        // check if local file is a directory or if fileSize match
                        fileInfo.setStatus( FileInfo::Success, copy.file().exists() && ( copy.file().isDirectory() || copy.file().fileSize() == fileInfo.size() ) );
                        copy.update();

                        updatedFiles.append( copy );

                    } else {

                        fileInfo.setFailed();
                        remotePathSet.insert( copy.file().path( false ) );

                    }

                }

                // emit fileInfo update signal
                Debug::Throw() << "Ftp::Interface::_runMacro - files: " << updatedFiles.size() << endl;
                if( !updatedFiles.isEmpty() ) emit fileInfoUpdated( updatedFiles );

                // if no local files are found (to check against remote), the macro is finished
                if( remotePathSet.isEmpty() ) _macroCompleted();
                else {

                    for( const auto& file:remotePathSet )
                    { _addCommand( Command( Command::Type::ShowDirectory, { FileInfo( file, FileInfo::Folder | FileInfo::Remote ) } ) ); }

                }

            }
            break;

            // checkFiles macro
            case Macro::Type::CheckFilesExist:
            {

                // retrieve local and remote path
                const QString buffer( tr( "Checking files existence\n" ) );
                emit logAvailable( buffer, Format::Bold );

                // keep track of path to list for conflicts
                QOrderedSet<File> remotePathSet;

                // loop over recursive files
                for( auto& fileInfo:filesRecursive_ )
                {

                    // retrieve associated file
                    // check location
                    if( fileInfo.isLocal() )
                    {

                        // check file exists
                        fileInfo.setStatus( FileInfo::Success, fileInfo.file().exists() );
                        fileInfo.update();

                    } else if( fileInfo.file().get() == "/" ) {

                        fileInfo.setSuccess();

                    } else {

                        fileInfo.setFailed();
                        Debug::Throw() << "Ftp::Interface::_runMacro - checking: " << fileInfo.file() << endl;
                        remotePathSet.insert( fileInfo.file().path( false ) );

                    }

                }

                // if no local files are found (to check against remote), the macro is finished
                if( remotePathSet.isEmpty() ) _macroCompleted();
                else {

                    for( const auto& file:remotePathSet )
                    {
                        Debug::Throw() << "Ftp::Interface::_runMacro - adding: " << file << endl;
                        _addCommand( Command( Command::Type::ShowDirectory, { FileInfo( file, FileInfo::Folder | FileInfo::Remote ) } ) );
                    }

                }

            }
            break;

            default:
            _macroCompleted();
            break;

        }

        return;

    }
    //_________________________________________________________________
    void Interface::_macroCompleted( bool error )
    {

        Debug::Throw( "Ftp::Interface::_macroCompleted.\n" );
        if( macros_.empty() ) return;

        // copy current macro
        Macro macro( macros_.front() );

        switch( macro.type() )
        {

            case Macro::Type::ShowDirectory:
            {
                emit contentAvailable( macro.argument(0), filesRecursive_ );
                break;
            }

            case Macro::Type::SizeRecursive:
            {
                if( macro.argument(0).isRemote() )
                {
                    _computeTotalSize();
                    emit sizeAvailable( macro.argument(0), total_ );
                }
                break;
            }

            case Macro::Type::CheckFilesCopy:
            case Macro::Type::CheckFilesMoved:
            {

                // loop over recursiveFiles
                for( const auto& fileInfo:filesRecursive_ )
                {
                    if( fileInfo.isSuccess() ) continue;
                    const QString buffer = tr( "Failed to transfer file '%1'\n" ).arg( fileInfo.file() );
                    emit logAvailable( buffer, Format::Italic );
                }

            }
            break;

            case Macro::Type::Open:
            {
                // retrieve local and remote path
                File localPath( macro.argument(0).file() );
                File remotePath( macro.argument(1).file() );

                // loop over recursive files
                FileInfo::ListIterator iter( filesRecursive_ );
                iter.toBack();
                while( iter.hasPrevious() )
                {
                    auto fileInfo( iter.previous() );
                    auto file( fileInfo.file() );

                    // check file type
                    if( !( fileInfo.isRemote() && fileInfo.isDocument() ) ) continue;

                    // create local temporary file file
                    _removePath( file, remotePath );
                    File tmpFile( Util::tmp() + "/qftp/" + Util::user() + file.addPath( remotePath, false ) );
                    if( !tmpFile.exists() )
                    {

                        // file does not exist
                        const QString buffer = tr( "Failed to transfer file '%1'\n" ).arg( fileInfo.file() );
                        emit logAvailable( buffer, Format::Italic );

                    } else {

                        // notify scratch file creation
                        emit scratchFileCreated( tmpFile );

                        // open file using specified command
                        if( macro.hasArgument(2) )
                        {
                            const File command( macro.argument(2).file() );
                            Base::Command( { command, tmpFile } ).run( tmpFile.path() );
                        } else {

                            // use default application otherwise
                            QDesktopServices::openUrl( QUrl::fromEncoded( QString( "file://%1" ).arg( tmpFile ).toLatin1() ) );

                        }
                    }
                }

            }
            break;

            default:
            break;
        }

        if( error ) {

            const QString buffer = tr( "Error while processing macro '%1'\n" ).arg( macro.name() );
            emit logAvailable( buffer, Format::Italic );

        }

        emit logAvailable( "\n" );
        emit macroCompleted( macro, error );
        macros_.removeFirst();
        _runMacro();

    }

    //_________________________________________________________________
    void Interface::_commandStarted( int id )
    {

        if( commands_.empty() ) return;
        Q_ASSERT( commands_.front().ftpId() == id );
        emit commandStarted( commands_.front() );

    }

    //_________________________________________________________________
    void Interface::_commandCompleted( int id, bool error )
    {

        // process next command
        if( commands_.isEmpty() )
        {
            _macroCompleted();
            return;
        }

        // store command and remove from map
        Command command( commands_.front() );
        Q_ASSERT( command.ftpId() == id );

        // report command status
        _logCommand( command, error );

        // report additional error message
        if( error && !ftp_->errorString().isEmpty() )
        {
            QString buffer;
            QTextStream( &buffer ) << ftp_->errorString() << endl;
            emit logAvailable( buffer, Format::Italic );

            ErrorDialog dialog( qApp->activeWindow(), ftp_->errorString() );
            if( dialog.exec() == QDialog::Accepted )
            {

                clear();
                ftp_->abort();

                // release busy state and exit
                QMetaObject::invokeMethod( this, "ready", Qt::QueuedConnection );
                emit busyStateChanged( false );
                return;

            }
        }

        // retrieve current macro (if any)
        Macro macro( macros_.isEmpty() ? Macro():macros_.front() );

        // process command
        switch (command.type() )
        {

            // for list command, add command path to found items
            case Command::Type::ShowDirectory:
            {

                // do nothing for local command
                if( command.argument(0).isLocal() ) break;

                // process directory
                // add current url list
                for( const auto& fileInfo:files_ )
                {

                    // create remote file info
                    FileInfo remoteFileInfo( fileInfo );
                    remoteFileInfo.setFile( File( fileInfo.file() ).addPath( command.argument(0).file(), false ) );

                    if(
                        macro.type() == Macro::Type::CheckConflicts ||
                        macro.type() == Macro::Type::CheckFilesCopy ||
                        macro.type() == Macro::Type::CheckFilesMoved )
                    {

                        // store source and destination path
                        auto sourcePathInfo( macro.argument(0) );
                        auto destinationPathInfo( macro.argument(1) );

                        // create matching local file
                        File localFile( remoteFileInfo.file() );
                        _removePath( localFile, destinationPathInfo.file() );
                        localFile.addPath( sourcePathInfo.file() );

                        auto localFileIter = std::find_if( filesRecursive_.begin(), filesRecursive_.end(), FileInfo::SameFileFTor(localFile) );
                        if( localFileIter == filesRecursive_.end() ) continue;
                        else if( macro.type() == Macro::Type::CheckConflicts )
                        {

                            // local file conflicts
                            localFileIter->setStatus( FileInfo::Conflict, true );

                            if( !localFileIter->isFolder() )
                            {
                                Debug::Throw() << "Ftp::Interface::_commandCompleted -"
                                    << " remote file: " << remoteFileInfo.file()
                                    << " last modified: " << remoteFileInfo.lastModified().toString()
                                    << " local file: " << localFile
                                    << " last modified: " << localFileIter->lastModified().toString()
                                    << endl;
                            }

                            // by default, skip file if it is a directory or if it is older
                            if( localFileIter->isFolder() || localFileIter->isOlder( remoteFileInfo.lastModified() ) )
                            { localFileIter->setStatus( FileInfo::Overwrite, false ); }

                        } else if( macro.type() == Macro::Type::CheckFilesCopy || macro.type() == Macro::Type::CheckFilesMoved ) {

                            // check if matching remote file is a directory or if fileSize match
                            localFileIter->setStatus( FileInfo::Success, localFileIter != filesRecursive_.end() && ( remoteFileInfo.isFolder() || localFileIter->size() == remoteFileInfo.size() ) );

                            if( !localFileIter->isFolder() )
                            {
                                Debug::Throw() << "Ftp::Interface::_commandCompleted -"
                                    << " remote file: " << remoteFileInfo.file()
                                    << " last modified: " << remoteFileInfo.lastModified().toString()
                                    << " local file: " << localFile
                                    << " last modified: " << localFileIter->lastModified().toString()
                                    << endl;
                            }

                            // emit file as updated
                            emit fileInfoUpdated( remoteFileInfo );

                        }

                    } else if( macro.type() == Macro::Type::CheckFilesExist ) {

                        Debug::Throw() << "Ftp::Interface::_commandCompleted - found " << remoteFileInfo.file() << endl;
                        auto fileIter = std::find_if( filesRecursive_.begin(), filesRecursive_.end(), FileInfo::SameFileFTor( remoteFileInfo.file() ) );
                        if( fileIter != filesRecursive_.end() ) fileIter->setSuccess();

                    } else if( macro.type() == Macro::Type::ShowDirectory && !_showHiddenFiles() && remoteFileInfo.file().isHidden() ) {

                        // skip file if hidden
                        continue;

                    } else {

                        // insert file into recursive file list
                        _addFileRecursive( remoteFileInfo );

                        // emit message
                        if( !error )
                        {
                            auto buffer = tr( "Adding file '%1' to cache\n" ).arg( remoteFileInfo.file() );
                            emit logAvailable( buffer );
                        }

                        // list subdirectories in case of list recursive command
                        if( ( macro.type() == Macro::Type::ListRecursive || macro.type() == Macro::Type::SizeRecursive ) && remoteFileInfo.isFolder() )
                        {
                            _addCommand( Command( Command::Type::ChangeDirectory, { remoteFileInfo } ) );
                            _addCommand( Command( Command::Type::ShowDirectory, { remoteFileInfo } ) );
                        }

                    }
                }

                // update total size
                if( macro.type() == Macro::Type::SizeRecursive )
                {
                    if( _updateTotalSize() )
                    { emit sizeAvailable( macro.argument(0), total_ ); }
                }

                // clear files
                _clearFiles();

            }
            break;

            case Command::Type::Copy:
            {

                // close stream
                if( ftp_->currentDevice() )
                {
                    ftp_->currentDevice()->close();
                    delete ftp_->currentDevice();
                }

                // update transfer size
                if( macro.type() == Macro::Type::CopyRecursive )
                {
                    _updateTransferedSize( command.argument(0) );
                    emit progressAvailable( transfered_ );
                }

            }
            break;

            case Command::Type::CreateDirectory:
            if( macro.type() == Macro::Type::CopyRecursive )
            {
                _updateTransferedSize( command.argument(0) );
                emit progressAvailable( transfered_ );
            }
            break;

            case Command::Type::Remove:
            if( macro.type() == Macro::Type::Remove )
            {
                _updateTransferedSize( command.argument(0) );
                emit progressAvailable( transfered_ );
            }
            break;

            default: break;
        }

        // emit signal, remove from pile and run next command if any
        emit commandCompleted( command, error );
        commands_.removeFirst();
        _runCommand();

        return;

    }

    //_________________________________________________________________
    void Interface::_progress( qint64 done, qint64 )
    {
        Debug::Throw( "Ftp::Interface::_progress.\n" );
        emit progressAvailable( transfered_ + done );
    }

    //___________________________________________________________________________
    void Interface::_addRemoteFile( const QUrlInfo& file )
    {

        Debug::Throw( "Ftp::Interface::_addRemoteFile.\n" );

        // ignore navigation
        if( file.name() == "." || file.name() == ".." ) return;

        // convert to FileInfo
        FileInfo fileInfo( file.name(), FileInfo::Remote );

        // set type
        if( file.isDir() ) fileInfo.setIsFolder();
        else fileInfo.setIsDocument();
        if( file.isSymLink() ) fileInfo.setIsLink();

        // set size and modification
        fileInfo.setSize( file.size() );
        fileInfo.setLastModified( TimeStamp( file.lastModified().toTime_t() ) );
        fileInfo.setUser( file.owner() );
        fileInfo.setGroup( file.group() );

        // store
        _addFile( fileInfo );

    }

    //___________________________________________________________________________
    void Interface::_processFiles( const File::List& files )
    {

        // get current macro
        if( macros_.empty() ) return;
        Macro macro( macros_.front() );

        switch( macro.type() )
        {

            case Macro::Type::ShowDirectory:
            case Macro::Type::ListRecursive:
            for( const auto& file:files )
            {

                // check if hidden
                if( !_showHiddenFiles() && file.isHidden() ) continue;

                // propagate file properties to FileInfo
                FileInfo fileInfo( file );
                if( file.isDirectory() ) fileInfo.setIsFolder();
                else fileInfo.setIsDocument();

                if( file.isLink() ) fileInfo.setIsLink();
                if( file.isBrokenLink() ) fileInfo.setIsBrokenLink();
                if( file.isHidden() ) fileInfo.setIsHidden();

                fileInfo.setSize( file.fileSize() );
                fileInfo.setLastModified( file.lastModified() );
                fileInfo.setUser( file.userName() );
                fileInfo.setGroup( file.groupName() );
                fileInfo.setPermissions( file.permissions() );

                _addFileRecursive( fileInfo );

                const QString buffer = tr( "Adding file '%1' to cache\n" ).arg( fileInfo.file() );
                emit logAvailable( buffer );
            }
            break;

            default: break;

        }
    }

    //___________________________________________________________________________
    void Interface::_processSize( qint64 size )
    {

        // get current macro
        if( macros_.empty() ) return;
        Macro macro( macros_.front() );
        if( macro.type() == Macro::Type::SizeRecursive )
        { emit sizeAvailable( macro.argument(0), size ); }

    }

    //_________________________________________________________________
    void Interface::_addCommand( const Command& command )
    {
        Debug::Throw() << "Ftp::Interface::_addCommand - " << command.name() << endl;

        // add command
        const bool empty( commands_.empty() );
        commands_.append( command );

        // emit signal
        emit commandAdded( command );

        // run command if there is none pending
        if( empty ) _runCommand();

    }


    //__________________________________________________________________________________
    void Interface::_addCommands( const Command::List& commands )
    {

        Debug::Throw( "Ftp::Interface::_addCommands.\n" );

        const bool empty( commands_.empty() );
        commands_.append( commands );

        // check if commands have been effectively added
        emit commandsAdded( commands );
        if( empty ) _runCommand();

    }

    //_________________________________________________________________
    void Interface::_runCommand()
    {

        // nothing if there is no command in the pipe
        if( commands_.isEmpty() )
        {
            _macroCompleted();
            return;
        }

        // store command and path
        Command& command( commands_.front() );
        Debug::Throw() << "Ftp::Interface::_runCommand - command: " << command.name() << " (" << command.id() << ")" << endl;

        // retrieve current macro (if any)
        Macro macro( macros_.isEmpty() ? Macro():macros_.front() );

        switch(command.type() )
        {

            case Command::Type::ConnectToHost:
            {
                QString host( command.argument(0).file() );
                quint16 port( command.argument(1).file().get().toUInt() );
                command.setFtpId( ftp_->connectToHost( host, port ) );
                return;
            }

            case Command::Type::Login:
            {
                QString user( command.argument(0).file() );
                QString password( command.argument(1).file() );
                command.setFtpId( ftp_->login( user, password ) );
                return;
            }

            case Command::Type::Close:
            {
                command.setFtpId( ftp_->close() );
                return;
            }

            case Command::Type::ChangeDirectory:
            {
                command.setFtpId( ftp_->cd( command.argument(0).file() ) );
                return;
            }

            case Command::Type::ShowDirectory:
            {
                const FileInfo& path( command.argument(0) );

                if( path.isLocal() )
                {

                     // check if path is valid
                    if( !( path.file().exists() && path.file().isDirectory() ) )
                    {

                        const QString buffer = tr( "Failed to change directory to '%1'\n" ).arg( path.file() );
                        emit logAvailable( buffer, Format::Italic );
                        break;

                    }

                    // setup thread
                    thread_->setFile( path.file() );

                    if( macro.type() == Macro::Type::ShowDirectory ) thread_->setCommand( FileThread::Command::List );
                    else if( macro.type() == Macro::Type::SizeRecursive ) thread_->setCommand( FileThread::Command::SizeRecursive );
                    else if( macro.type() == Macro::Type::ListRecursive ) thread_->setCommand( FileThread::Command::ListRecursive );

                    File::ListFlags flags( File::ListFlag::None );
                    if( macro.type() == Macro::Type::ListRecursive || macro.type() == Macro::Type::SizeRecursive )
                    { flags |= File::ListFlag::Recursive; }

                    if( _showHiddenFiles() || macro.type() == Macro::Type::SizeRecursive )
                    { flags |= File::ListFlag::ShowHiddenFiles; }

                    if( macro.type() == Macro::Type::ListRecursive && macro.hasArgument(1) && macro.argument(1).isRemote() )
                    { flags |= File::ListFlag::FollowLinks; }

                    thread_->setFlags( flags );
                    thread_->start();
                    return;

                } else {

                    command.setFtpId( ftp_->list( command.argument(0).file() ) );
                    return;

                }

            }

            case Command::Type::Copy:
            {

                const FileInfo& source( command.argument(0) );
                const FileInfo& destination( command.argument(1) );

                const QString buffer = tr( "Copying '%1' to '%2'\n" ).arg( source.file(), destination.file() );
                emit logAvailable( buffer );

                if( source.isLocal() && destination.isLocal() )
                {

                    // local to local copy
                    thread_->setFile( source.file() );
                    thread_->setDestination( destination.file() );
                    thread_->setCommand( FileThread::Command::Copy );
                    thread_->start();

                    emit commandStarted( command );
                    return;

                } else if( source.isLocal() ) {

                    // local to remote copy
                    QFile* in = new QFile( source.file() );
                    if( in->open( QIODevice::ReadOnly ) )
                    {

                        command.setFtpId( ftp_->put( in, destination.file() ) );
                        return;

                    } else {

                        // cannot read file
                        // emit error message and remove command from list
                        const QString buffer = tr( "Failed to read '%1'\n" ).arg( source.file() );
                        emit logAvailable( buffer, Format::Italic );
                        break;

                    }

                } else {

                    // remote to local copy
                    QFile* out = new QFile( destination.file() );
                    if( out->open( QIODevice::WriteOnly ) )
                    {

                        command.setFtpId( ftp_->get( source.file(), out ) );
                        return;

                    } else {

                        // cannot write to file
                        // emit error message and remove command from list
                        const QString buffer = tr( "Failed to write to '%1'\n" ).arg( destination.file() );
                        emit logAvailable( buffer, Format::Italic );
                        break;

                    }

                }

            }

            case Command::Type::Rename:
            {

                const FileInfo& source( command.argument(0) );
                const FileInfo& destination( command.argument(1) );

                if( source.isLocal() )
                {
                    if( !source.file().rename( destination.file() ) ) emit commandFailed( command );
                    break;

                } else {

                    command.setFtpId( ftp_->rename( source.file(), destination.file() ) );
                    return;

                }

            }

            case Command::Type::CreateDirectory:
            {
                const FileInfo& path( command.argument(0) );
                if( path.isLocal() )
                {
                    if( !path.file().createDirectory() )
                    {

                        const QString buffer = tr( "Failed to create folder '%1'\n" ).arg( path.file() );
                        emit logAvailable( buffer );
                        emit commandFailed( command, buffer );

                    } else emit fileInfoAdded( path );
                    break;

                } else {

                    command.setFtpId( ftp_->mkdir( path.file() ) );
                    return;
                }

            }

            case Command::Type::Remove:
            {
                const FileInfo& path( command.argument(0) );
                if( path.isLocal() )
                {

                    path.file().removeRecursive();
                    if( path.file().exists() ) {

                        const QString buffer = tr( "Failed to remove file '%1'\n" ).arg( path.file() );
                        emit logAvailable( buffer );
                        emit commandFailed( command );

                    } else emit fileInfoRemoved( path );
                    break;

                } else {

                    if( path.isFolder() ) command.setFtpId( ftp_->rmdir( path.file() ) );
                    else command.setFtpId( ftp_->remove( path.file() ) );
                    return;

                }
            }

            default: break;

        }

        emit commandCompleted( command, true );
        commands_.removeFirst();

        // run next command/macro
        if( commands_.isEmpty() ) _macroCompleted();
        else _runCommand();

        return;

    }

    //_________________________________________________________________________________________________
    void Interface::_logCommand( const Command& command, bool error )
    {

        Debug::Throw( "Ftp::Interface::_logCommand.\n" );
        if( !error ) return;

        QString buffer;
        switch( command.type() )
        {
            case Command::Type::CreateDirectory:
            {
                buffer = tr( "Failed to create directory '%1'\n" ).arg( command.argument(0).file() );
                break;
            }

            case Command::Type::Rename:
            {
                buffer = tr( "Failed to rename file '%1' to '%2'\n" ).arg( command.argument(0).file(), command.argument(1).file() );
                break;
            }

            case Command::Type::Copy:
            {
                buffer = tr( "Failed to copy '%1'\n" ).arg( command.argument(0).file() );
                break;
            }

            default: break;

        }

        if( !buffer.isEmpty() ) emit logAvailable( buffer, Format::Italic );
        return;


    }

    //_________________________________________________________________________________________________
    void Interface::_removePath( File& source, File path ) const
    {
        path.addTrailingSlash();
        if( !source.startsWith( path ) ) return;
        source = File( source.get().mid( path.get().size() ) );
    }

    //_________________________________________________________________
    void Interface::_resetTotalSize()
    {
        Debug::Throw( "Ftp::Interface::_resetTotalSize.\n" );

        // compute total size to transfer
        total_ = 1;
        transfered_ = 0;
    }

    //_________________________________________________________________
    bool Interface::_updateTotalSize()
    {

        Debug::Throw( "Ftp::Interface::_updateTotalSize.\n" );
        const auto total = std::accumulate( files_.begin(), files_.end(), qint64(0),
            []( qint64 total, const FileInfo& fileInfo ) { return total+fileInfo.size(); } );

        if( total >= 0 )
        {

            total_+= total;
            return true;

        } else return false;

    }

    //_________________________________________________________________
    void Interface::_computeTotalSize()
    {
        Debug::Throw( "Ftp::Interface::_computeTotalSize.\n" );

        // compute total size to transfer
        total_ = std::accumulate( filesRecursive_.begin(), filesRecursive_.end(), qint64(0),
            []( qint64 total, const FileInfo& fileInfo )
            { return fileInfo.isOverwrite() ? total + fileInfo.size() : total; } );

        transfered_ = 0;

        // needed because progress-bars start "oscillating"
        // when a total size of 1 is set.
        if( total_ <= 0 ) total_ = 1;

        return;
    }

}
