#ifndef FtpInterface_h
#define FtpInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"
#include "FileThread.h"
#include "FtpCommand.h"
#include "FtpMacro.h"
#include "TextFormat.h"

#include <QColor>
#include <QCloseEvent>
#include <QFtp>

class FileThread;

namespace Ftp
{

    //* interface to ftp commands
    class Interface: public QObject, private Base::Counter<Interface>
    {

        //* Qt meta object declaration
        Q_OBJECT

        public:

        //* constructor;
        explicit Interface( QObject* );

        //* clear
        void clear()
        {
            Debug::Throw( "Ftp::Interface::clear.\n" );
            macros_.clear();
            commands_.clear();
        }

        //* reference to ftp object
        QFtp& ftp() const
        { return *ftp_; }

        //*@name interface to store files
        //@{

        //* list of files
        const FileInfo::List& filesRecursive() const
        { return filesRecursive_; }

        //* list of files
        FileInfo::List& filesRecursive()
        { return filesRecursive_; }

        //@}

        Q_SIGNALS:

        //* directory content is available
        void contentAvailable( const FileInfo&, const FileInfo::List& );

        //* directory size
        void sizeAvailable( const FileInfo&, qint64 );

        //* emitted when some FileInfo has been updated
        void fileInfoUpdated( const FileInfo& );

        //* emitted when some FileInfo has been updated
        void fileInfoUpdated( const FileInfo::List& );

        //* file added
        void fileInfoAdded( const FileInfo& );

        //* file removed
        void fileInfoRemoved( const FileInfo& );

        //*@name macro interface
        //@{

        //* new macro added
        void macroAdded( const Ftp::Macro& );

        //* macro removed (prior to start)
        void macroRemoved( const Ftp::Macro& );

        //* new macro started
        void macroStarted( const Ftp::Macro& );

        //* macro completed
        void macroCompleted( const Ftp::Macro&, bool );

        //* ftp command added
        void commandAdded( const Ftp::Command& );

        //* commands added
        void commandsAdded( const Ftp::Command::List& );

        //* ftp command started
        void commandStarted( const Ftp::Command& );

        //* ftp command completed
        void commandCompleted( const Ftp::Command&, bool );

        //* command failed
        void commandFailed( const Ftp::Command&, const QString& = QString() );

        //@}

        //* emitted for logging of processes
        void logAvailable( const QString&, Format::TextFormatFlags = Format::Default, QColor = QColor(), int = 0 );

        //* emitted when interface is busy
        void busy();

        //* emitted when interface is idle
        void ready();

        //* emitted when interface is busy
        void busyStateChanged( bool );

        //* emitted to show transfer progress
        void maxProgressAvailable( int );

        //* emitted to show transfer progress
        void progressAvailable( int );

        //* emitted when scratch file is created
        void scratchFileCreated( const File& );

        public Q_SLOTS:

        //* add ftp macro
        void addMacro( const Ftp::Macro& );

        //* add ftp macro
        void addMacros( const Ftp::Macro::List& );

        //* show hidden files
        void setShowHiddenFiles( bool );

        private Q_SLOTS:

        //* run macro
        void _runMacro();

        //* current macro finished
        void _macroCompleted( bool = false );

        //* command finished
        void _commandStarted( int );

        //* command finished
        void _commandCompleted( int = 0, bool = false );

        //* transfer progress
        void _progress( qint64, qint64 );

        //* add files from remote listing
        void _addRemoteFile( const QUrlInfo& );

        //* process files from thread
        void _processFiles( const File::List& );

        //* process size from thread
        void _processSize( qint64 );

        private:

        //*@name command interface
        //@{

        //* add command to queue
        void _addCommand( const Command& );

        //* add commands
        void _addCommands( const Command::List& );

        //* process command from queue
        void _runCommand();

        //* emit message for given command and error status
        void _logCommand( const Command&, bool );

        //@}

        //* show hidden files
        bool _showHiddenFiles() const
        { return showHiddenFiles_; }

        //* add file to recursive list
        void _addFileRecursive( const FileInfo& file )
        { filesRecursive_ << file; }

        //* clear list of files
        void _clearFilesRecursive()
        { filesRecursive_.clear(); }

        //* add file
        void _addFile( const FileInfo& fileInfo )
        { files_ << fileInfo; }

        //* clear files
        void _clearFiles()
        { files_.clear(); }

        //* remove path from a given file
        void _removePath( File& source, File path ) const;

        //* reset total and transfered size
        void _resetTotalSize();

        //* update total size from file_ content
        bool _updateTotalSize();

        //* compute total size and reset transfered size
        void _computeTotalSize();

        //* update transfered size
        void _updateTransferedSize( const FileInfo& file )
        { transfered_ += file.size(); }

        //*@name stored url informations for remote listing
        //@{

        //* list of files
        /*!
        filesRecursive_ is used to store all files read either remotely
        or locally, that can be retrieved internally
        */
        FileInfo::List filesRecursive_;

        //* list of files
        /*!
        files_ store all files read from a single "ls" command,
        they are copied into filesRecursives_ when several such commands
        must be called in a row
        */
        FileInfo::List files_;

        //@}

        //* show hidden files
        bool showHiddenFiles_ = false;

        //* Ftp connection manager
        QFtp* ftp_ = nullptr;

        //* file thread, for local files manipulations
        FileThread* thread_ = nullptr;

        //* list of macros
        Macro::List macros_;

        //* map command ids and types
        Command::List commands_;

        //* total size to transfer
        qint64 total_ = 1;

        //* transfered size
        qint64 transfered_ = 0;

    };

}

#endif
