/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "FtpMacroModel.h"

#include "CustomPixmap.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "TimeStamp.h"
#include "XmlOptions.h"

#include <QIcon>

namespace Ftp
{
    //_______________________________________________________________
    MacroModel::MacroModel( QObject* parent ):
        ListModel( parent ),
        Counter( "Ftp::MacroModel" )
    {

        Debug::Throw( "Ftp::MacroModel::MacroModel.\n" );
        updateConfiguration();

    }

    //__________________________________________________________________
    QVariant MacroModel::data( const QModelIndex& index, int role ) const
    {

        // check index
        if( !contains( index ) ) return QVariant();

        // retrieve associated macro
        const Macro& macro( get()[index.row()] );

        // return text associated to macro and column
        if( role == Qt::DisplayRole ) {

            switch( index.column() )
            {

                case Name: return macro.name();
                case FileName: return macro.argumentCount() > 0  ? macro.argument(0).file().get():QVariant();
                default: return QVariant();
            }
        } else if( role == Qt::DecorationRole && index.column() == Name ) {

            // return icon associated to file
            return _icon(macro.type());

        }

        return QVariant();

    }

    //__________________________________________________________________
    QVariant MacroModel::headerData(int section, Qt::Orientation orientation, int role) const
    {

        if(
            orientation == Qt::Horizontal &&
            role == Qt::DisplayRole &&
            section >= 0 &&
            section < nColumns )
        { return columnTitles_[section]; }

        // return empty
        return QVariant();

    }

    //_________________________________________________________
    void MacroModel::updateConfiguration()
    {

        Debug::Throw( "MacroModel::updateConfiguration.\n" );

        beginResetModel();

        // reinitialize icons
        _icons().clear();
        _installIcons();

        endResetModel();

    }

    //_____________________________________________________________________
    void MacroModel::_installIcons() const
    {

        Debug::Throw( "MacroList::_installIcons" );

        if( _icons().empty() )
        {
            _icons() = Base::makeT<IconCache>( {
                { Macro::Type::Connect, IconEngine::get( IconNames::Connect ) },
                { Macro::Type::Add, IconEngine::get( IconNames::Add ) },
                { Macro::Type::Clear, IconEngine::get( IconNames::Delete ) },
                { Macro::Type::Abort, IconEngine::get( IconNames::Abort ) },
                { Macro::Type::CopyRecursive, IconEngine::get( IconNames::Paste ) },
                { Macro::Type::CreateDirectory, IconEngine::get( IconNames::CreateDirectory ) },
                { Macro::Type::Rename, IconEngine::get( IconNames::Rename ) },
                { Macro::Type::Remove, IconEngine::get( IconNames::Delete ) },
                { Macro::Type::Open, IconEngine::get( IconNames::Edit ) },
                { Macro::Type::ShowDirectory, IconEngine::get( IconNames::Find ) },
                { Macro::Type::ListRecursive, IconEngine::get( IconNames::Find ) }
            });
        }
        return;
    }

    //_______________________________________________
    MacroModel::IconCache& MacroModel::_icons()
    {
        static IconCache cache;
        return cache;
    }

    //____________________________________________________________
    const QIcon& MacroModel::_icon( Macro::Type type )
    {

        IconCache::iterator iter = ( _icons().find( type ) );
        if( iter != _icons().end() ) return iter.value();
        else return _icons().insert( type, CustomPixmap( IconSize::get( IconSize::Medium ), CustomPixmap::Flag::Transparent ) ).value();

    }
}
