/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/
#include "MainWindow.h"

#include "Application.h"
#include "BaseFindWidget.h"
#include "CheckFilesDialog.h"
#include "ColumnSelectionMenu.h"
#include "ColumnSortingMenu.h"
#include "ConflictDialog.h"
#include "ConnectionDialog.h"
#include "ConnectionFrame.h"
#include "ContextMenu.h"
#include "CopyFilesDialog.h"
#include "CppUtil.h"
#include "CustomComboBox.h"
#include "CustomToolBar.h"
#include "DockWidget.h"
#include "FileIconProvider.h"
#include "FileInformationDialog.h"
#include "FileSelectionFrame.h"
#include "FtpCommandDialog.h"
#include "FtpDefines.h"
#include "FtpInterface.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "LogDialog.h"
#include "LogWidget.h"
#include "MenuBar.h"
#include "OpenWithDialog.h"
#include "PathEditor.h"
#include "PlacesWidget.h"
#include "ProgressStatusBar.h"
#include "QuestionDialog.h"
#include "Singleton.h"
#include "TaskBarProgressNotifications.h"
#include "TreeView.h"
#include "Util.h"
#include "ViewModeMenu.h"
#include "XmlOptions.h"

#include <QDir>
#include <QApplication>
#include <QLayout>
#include <QSplitter>

#include <numeric>

//____________________________________________________________________________________________
MainWindow::MainWindow( QWidget* parent ):
    BaseMainWindow( parent ),
    Counter( "MainWindow" ),
    fileSystemWatcher_( this ),
    interface_( new Ftp::Interface( this ) )
{

    Debug::Throw( "MainWindow::MainWindow.\n" );
    setOptionName( "MAIN_WINDOW" );

    // actions
    _installActions();
    auto application( Base::Singleton::get().application<Application>() );
    addAction( &application->closeAction() );

    // main widget and layout
    auto main = new QWidget( this );
    auto mainLayout( new QVBoxLayout );
    main->setLayout( mainLayout );
    mainLayout->setMargin(0);
    mainLayout->setSpacing(2);
    setCentralWidget( main );

    auto hLayout( new QHBoxLayout );
    hLayout->setMargin(0);
    hLayout->setSpacing(2);
    mainLayout->addLayout( hLayout );

    // right container (for find widget)
    rightContainer_ = new QWidget( main );
    rightContainer_->setLayout( new QVBoxLayout );
    rightContainer_->layout()->setMargin(0);
    rightContainer_->layout()->setSpacing(0);
    hLayout->addWidget( rightContainer_ );

    // splitter
    auto splitter = new QSplitter( rightContainer_ );
    static_cast<QBoxLayout*>(rightContainer_->layout())->addWidget( splitter, 1 );

    // file selection frames
    splitter->addWidget( localSelectionFrame_ = new FileSelectionFrame( splitter, FileInfo::None ) );
    splitter->addWidget( remoteSelectionFrame_ = new FileSelectionFrame( splitter, FileInfo::Remote ) );

    // create icon provider and assign
    iconProvider_ = new FileIconProvider( this );
    for( const auto& frame:selectionFrames() )
    { frame->model().setIconProvider( iconProvider_ ); }

    //* dialogs
    commandDialog_ = new Ftp::CommandDialog( this );
    logDialog_ = new LogDialog( this );

    connect( commandDialog_, SIGNAL(dialogHidden()), SLOT(_commandDialogHidden()) );
    connect( logDialog_, SIGNAL(dialogHidden()), SLOT(_logDialogHidden()) );

    // toolbars
    {
        auto toolbar = new CustomToolBar( tr( "Main Toolbar" ), this, "TRANSFER_TOOLBAR" );
        toolbar->addAction( toggleConnectionAction_ );
        toolbar->addSeparator();
        toolbar->addAction( homeDirectoryAction_ );
        toolbar->addWidget( previousDirectoryButton_ );
        toolbar->addWidget( nextDirectoryButton_ );
        toolbar->addAction( parentDirectoryAction_ );
        toolbar->addSeparator();
        toolbar->addAction( reloadAction_ );
        toolbar->addAction( abortAction_ );
    }

    {
        auto toolbar =  new CustomToolBar( tr( "View Modes" ), this, "VIEW_MODES_TOOLBAR" );
        toolbar->addAction( iconsViewAction_ );
        toolbar->addAction( detailsViewAction_ );
    }

    // dock widgets
    placesPanel_ = new DockWidget( tr( "Places" ), this, "PLACES" );
    placesPanel_->setAllowedAreas( Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea );
    placesPanel_->setUseScrollArea( true );
    placesPanel_->visibilityAction().setText( tr( "Show Favorite Places" ) );
    addDockWidget( Qt::LeftDockWidgetArea, placesPanel_ );

    placesWidget_ = new PlacesWidget;
    placesWidget_->setIconProvider( iconProvider_ );
    placesPanel_->setMainWidget( placesWidget_ );

    if( !placesPanel_->visibilityAction().isChecked() )
    { placesPanel_->hide(); }

    // find widget
    _createFindWidget();

    // status bar
    setStatusBar( statusbar_ = new ProgressStatusBar( this ) );
    statusbar_->addClock();

    // menu bar
    auto menuBar = new MenuBar( this );
    setMenuBar( menuBar );
    connect( &menuBar->viewModeMenu(), SIGNAL(viewModeSelected(FileSelectionFrame::ViewMode)), SLOT(_updateViewMode(FileSelectionFrame::ViewMode)) );

    // interface connection
    connect( interface_, SIGNAL(contentAvailable(FileInfo,const FileInfo::List&)), SLOT(_updateContent(FileInfo,const FileInfo::List&)) );
    connect( interface_, SIGNAL(macroStarted(Ftp::Macro)), SLOT(_macroStarted(Ftp::Macro)) );
    connect( interface_, SIGNAL(macroCompleted(Ftp::Macro,bool)), SLOT(_macroCompleted(Ftp::Macro,bool)) );
    connect( interface_, SIGNAL(commandCompleted(Ftp::Command,bool)), SLOT(_commandCompleted(Ftp::Command,bool)) );
    connect( interface_, SIGNAL(fileInfoUpdated(FileInfo)), SLOT(_updateFileInfo(FileInfo)) );
    connect( interface_, SIGNAL(fileInfoUpdated(const FileInfo::List&)), SLOT(_updateFileInfo(const FileInfo::List&)) );

    connect( interface_, SIGNAL(maxProgressAvailable(int)), &statusbar_->progressBar(), SLOT(setMaximum(int)) );
    connect( interface_, SIGNAL(progressAvailable(int)), &statusbar_->progressBar(), SLOT(setValue(int)) );

    taskBarProgressNotifications_ = new TaskBarProgressNotifications( this );
    taskBarProgressNotifications_->setApplicationName( "qftp" );
    connect( interface_, SIGNAL(maxProgressAvailable(int)), taskBarProgressNotifications_, SLOT(setMaximum(int)) );
    connect( interface_, SIGNAL(progressAvailable(int)), taskBarProgressNotifications_, SLOT(setValue(int)) );

    connect( interface_, SIGNAL(commandAdded(Ftp::Command)), &commandDialog_->commandModel(), SLOT(add(Ftp::Command)) );
    connect( interface_, SIGNAL(commandsAdded(Ftp::Command::List)), &commandDialog_->commandModel(), SLOT(add(Ftp::Command::List)) );
    connect( interface_, SIGNAL(commandCompleted(Ftp::Command,bool)), &commandDialog_->commandModel(), SLOT(remove(Ftp::Command)) );

    connect( interface_, SIGNAL(macroAdded(Ftp::Macro)), &commandDialog_->macroModel(), SLOT(add(Ftp::Macro)) );
    connect( interface_, SIGNAL(macroCompleted(Ftp::Macro,bool)), &commandDialog_->macroModel(), SLOT(remove(Ftp::Macro)) );

    connect( interface_, SIGNAL(busyStateChanged(bool)), SLOT(_updateBusyButtons(bool)) );
    connect( interface_, SIGNAL(busyStateChanged(bool)), &abortAction(), SLOT(setEnabled(bool)) );
    connect( &interface_->ftp(), SIGNAL(stateChanged(int)), SLOT(_stateChanged(int)) );
    connect( interface_, SIGNAL(logAvailable(QString,Format::TextFormatFlags,QColor,int)), &logDialog().editor(), SLOT(append(QString,Format::TextFormatFlags,QColor,int)) );

    connect( this, SIGNAL(macroRequested(Ftp::Macro)), interface_, SLOT(addMacro(Ftp::Macro)) );
    connect( this, SIGNAL(macrosRequested(Ftp::Macro::List)), interface_, SLOT(addMacros(Ftp::Macro::List)) );

    for( const auto& frame:selectionFrames() )
    {

        connect( frame, SIGNAL(fileSelected(FileInfo)), SLOT(_fileSelected(FileInfo)) );
        connect( frame, SIGNAL(selectionChanged()), SLOT(_updateStatusBarDelayed()) );
        connect( frame, SIGNAL(actionsChanged()), SLOT(_updateNavigationActions()) );
        connect( frame, SIGNAL(clipboardChanged(FileSelectionClipboard)), SLOT(_updateClipboard(FileSelectionClipboard)) );
        connect( frame, SIGNAL(pasteClipboardRequested(FileInfo)), SLOT(_pasteClipboard(FileInfo)) );
        connect( frame, SIGNAL(openFilesRequested(const FileInfo::List&)), SLOT(_openFiles(const FileInfo::List&)) );

        connect( frame, SIGNAL(macroRequested(Ftp::Macro)), interface_, SLOT(addMacro(Ftp::Macro)) );
        connect( frame, SIGNAL(macrosRequested(Ftp::Macro::List)), interface_, SLOT(addMacros(Ftp::Macro::List)) );
        connect( frame, SIGNAL(requestFileInformation(FileInfo,QIcon)), SLOT(_fileInformation(FileInfo,QIcon)) );

        connect( &frame->model(), SIGNAL(transferRequested(FileInfo,const FileInfo::List&)), SLOT(_pasteSelection(FileInfo,const FileInfo::List&)) );
        connect( &frame->visibilityAction(), SIGNAL(toggled(bool)), SLOT(_updateVisibilityActions()) );
        connect( frame->list().selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    }

    // places
    connect( placesWidget_, SIGNAL(itemSelected(BaseFileInfo)), SLOT(_placeSelected(BaseFileInfo)) );
    connect( placesWidget_, SIGNAL(remoteItemAdded(BaseFileInfo)), SLOT(_requestPlacesUpdate(BaseFileInfo)) );

    // connections
    connect( &fileSystemWatcher_, SIGNAL(directoryChangedDelayed(QString)), &localSelectionFrame_->reloadAction(), SLOT(trigger()) );

    // active frame change
    connect( qApp, SIGNAL(focusChanged(QWidget*,QWidget*)), SLOT(_updateActiveFrame(QWidget*,QWidget*)) );

    connect( &showHiddenFilesAction(), SIGNAL(toggled(bool)), SLOT(_toggleShowHiddenFiles(bool)) );

    // clipboard
    _updateClipboard();

    // configuration
    connect( application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

    // buttons
    _updateButtons();
    _updateBusyButtons( false );

    // initialize local home path
    localSelectionFrame_->setHomePath( Util::home() );

    // initialize local root path list
    File::List rootPathList;
    const auto drives( QDir::drives() );
    std::transform( drives.begin(), drives.end(), std::back_inserter( rootPathList ),
        []( const QFileInfo& fileInfo ) { return File( fileInfo.path() ); } );

    localSelectionFrame_->setRootPathList( rootPathList );

    // initialize local path to working directory
    File path( Util::workingDirectory() );
    localSelectionFrame_->setPath( path );
    _fileSelected( FileInfo( path, FileInfo::Folder ) );

    // initialize remote home directory
    remoteSelectionFrame_->setHomePath( File( "/" ) );

    // reset remote file
    remoteSelectionFrame_->clear();
    remoteSelectionFrame_->setEnabled( false );

    // disable remote items in places widget, untill connected
    _disableRemotePlaces();

    // busy widget
    busyWidget_ = new BusyWidget( main, BusyWidget::Location::BottomRight );
    busyWidget_->hide();

}

//____________________________________________________________________________________________
FileSelectionFrame& MainWindow::activeFrame() const
{ return *(localSelectionFrame_->isActive() ? localSelectionFrame_:remoteSelectionFrame_); }

//____________________________________________________________________________________________
void MainWindow::connectToHost( QString host, int port, QString user, QString password, bool autoConnect )
{
    Debug::Throw( "MainWindow::connectToHost.\n" );

    ConnectionAttributes attributes;
    attributes.setHost( host );
    attributes.setPort( port );
    attributes.setUser( user );
    attributes.setPassword( password );

    if( !autoConnect )
    {
        ConnectionDialog dialog( this );
        dialog.frame().load( attributes );
        dialog.centerOnParent();

        if( !dialog.centerOnParent().exec() ) return;

        // update attributes from dialog
        attributes_ = dialog.frame().attributes();

    } else attributes_ = attributes;

    // disconnect if already connected
    if( interface_->ftp().state() == QFtp::LoggedIn ) emit macroRequested( Ftp::Macro( Ftp::Macro::Type::CloseConnection ) );

    // try connect
    emit macroRequested( Ftp::Macro( Ftp::Macro::Type::Connect,
    {
        FileInfo( attributes_.host() ),
        FileInfo( QString::number( attributes_.port() ) ),
        FileInfo( attributes_.user() ),
        FileInfo( attributes_.password() )
    } ) );
}


//_____________________________________________________________________
void MainWindow::findFromDialog()
{
    Debug::Throw( "MainWindow::findFromDialog.\n" );

    // create find widget
    if( !findWidget_ ) _createFindWidget();
    findWidget_->show();
    findWidget_->editor().setFocus();

    /*
    setting the default text values
    must be done after the dialog is shown
    otherwise it may be automatically resized
    to very large sizes due to the input text
    */
    QString text( activeFrame().textSelection().text() );
    if( !text.isEmpty() )
    {
        const int maxLength( 1024 );
        text = text.left( maxLength );
    }

    findWidget_->enableRegExp( true );
    findWidget_->synchronize();
    findWidget_->matchFound();

    return;
}

//____________________________________
void MainWindow::closeEvent( QCloseEvent* event )
{
    Debug::Throw( "MainWindow::closeEvent.\n" );
    event->accept();
    qApp->quit();
}

//____________________________________
void MainWindow::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == statusbarUpdateTimer_.timerId() )
    {

        statusbarUpdateTimer_.stop();
        _updateStatusBar();

    } else return BaseMainWindow::timerEvent( event );
}

//_______________________________________________________________________
bool MainWindow::_checkTransfer( const FileInfo& destination, const FileInfo::List& selection )
{

    Debug::Throw()
        << "MainWindow::_checkTransfer -"
        << " destination: " << destination.file()
        << " path: " << selectionFrame( destination.location() ).path().file()
        << endl;

    // make sure we are not copying file to same location
    if( destination.file() == selectionFrame( destination.location() ).path().file() )
    {

        if( std::any_of( selection.begin(), selection.end(),
            [&destination]( const FileInfo& fileInfo )
            {
                QString fileInfoPath( fileInfo.path().removeTrailingSlash() );
                return fileInfo.location() == destination.location() && fileInfoPath == destination.file();
            } ) )
        {
            InformationDialog dialog( this, tr( "Items cannot be copied onto themselves. <copy> aborted." ) );
            dialog.setWindowTitle( tr( "Warning" ) );
            dialog.exec();
            return false;
        }

    }

    // make sure one does not copy one file into itself
    if( destination.isFolder() )
    {
        if( std::any_of( selection.begin(), selection.end(),
            [&destination](const FileInfo& fileInfo )
            { return fileInfo.location() == destination.location() && fileInfo.file() == destination.file(); } ) )
        {

            InformationDialog dialog( this, tr( "A folder cannot be copied into itself. <copy> aborted." ) );
            dialog.setWindowTitle( tr( "Warning" ) );
            dialog.exec();
            return false;

        }

    }

    return true;

}

//____________________________________________________________________________________________
void MainWindow::_paste( const FileInfo& destination, const FileInfo::List& selection )
{

    Debug::Throw( "MainWindow::_paste.\n" );

    // retrieve selected files
    if( selection.empty() ) return;

    // check if copu is local to local
    const FileInfo sourcePath( selection.front().path(), FileInfo::Folder|selection.front().location() );
    const bool localOnly( sourcePath.isLocal() && destination.isLocal() );

    // check transfer types
    // allowed are local to local, local to remote and remote to local
    if( !( localOnly || interface_->ftp().state() == QFtp::LoggedIn ) ) return;

    // add list recursive commands
    Ftp::Macro::List macros;
    macros.append( Ftp::Macro( Ftp::Macro::Type::Clear ) );

    // add ListRecursive macros
    std::transform( selection.begin(), selection.end(), std::back_inserter( macros ),
        [&destination]( const FileInfo& fileInfo )
        { return Ftp::Macro( Ftp::Macro::Type::ListRecursive, { fileInfo, destination } ); } );

    // check conflicts, copy, and check result
    // update destination
    macros.append( Base::makeT<Ftp::Macro::List>(
    {
        Ftp::Macro( Ftp::Macro::Type::CheckConflicts, { sourcePath, destination } ),
        Ftp::Macro( Ftp::Macro::Type::CopyRecursive, { sourcePath, destination } ),
        Ftp::Macro( Ftp::Macro::Type::CheckFilesCopy, { sourcePath, destination } ),
        Ftp::Macro( Ftp::Macro::Type::ClearSelection, { destination } ),
        Ftp::Macro( Ftp::Macro::Type::Clear ),
        Ftp::Macro( Ftp::Macro::Type::ShowDirectory, { selectionFrame( destination.type() ).path() } )
    }) );

    // add selection macros
    const auto destinationList( _destinationList( destination, selection ) );
    std::transform( destinationList.begin(), destinationList.end(), std::back_inserter( macros ),
        []( const FileInfo& fileInfo )
        { return Ftp::Macro( Ftp::Macro::Type::Select, { fileInfo } ); } );

    emit macrosRequested( macros );

}

//_______________________________________________________________________
FileSelectionClipboard MainWindow::_move( const FileInfo& destination, const FileInfo::List& selection )
{

    Debug::Throw( "MainWindow::_move.\n" );

    // list of macros
    Ftp::Macro::List macros;

    // for local-to-local or remote-to-remote move, use Rename command
    if( selection.front().location() == destination.location() )
    {

        // loop over files in clipboard
        std::transform( selection.begin(), selection.end(), std::back_inserter( macros ),
            [&destination]( const FileInfo& fileInfo )
            {
                // generate destination file
                FileInfo destinationFile( fileInfo );
                destinationFile.setFile( fileInfo.file().localName().addPath( destination.file() ) );
                if( destination.isLocal() ) destinationFile.setLocal();
                else destinationFile.setRemote();
                return Ftp::Macro( Ftp::Macro::Type::Rename, { fileInfo, destinationFile } );
            } );

    } else {

        macros.append( Ftp::Macro( Ftp::Macro::Type::Clear ) );
        std::transform( selection.begin(), selection.end(), std::back_inserter( macros ),
            [&destination]( const FileInfo& fileInfo )
            { return Ftp::Macro( Ftp::Macro::Type::ListRecursive, { fileInfo, destination } ); } );

        // emit copy command
        const FileInfo sourcePath( selection.front().path(), FileInfo::Folder|selection.front().location() );

        // checkConflicts, copy, checkFiles and remove macros
        // TODO: make sure it does not remove files that have not been properly copied
        macros.append( Base::makeT<Ftp::Macro::List>(
        {
            Ftp::Macro( Ftp::Macro::Type::CheckConflicts, { sourcePath, destination } ),
            Ftp::Macro( Ftp::Macro::Type::CopyRecursive, { sourcePath, destination } ),
            Ftp::Macro( Ftp::Macro::Type::CheckFilesMoved, { sourcePath, destination } ),
            Ftp::Macro( Ftp::Macro::Type::Remove, { sourcePath } )
        }));

    }

    // generate updated clipboard
    FileSelectionClipboard localClipboard( _destinationList( destination, selection ) );
    localClipboard.setMode( FileSelectionClipboard::Copy );

    // trigger destination update
    if( destination == selectionFrame( destination.type() ).path() )
    {
        macros.append( Base::makeT<Ftp::Macro::List>(
        {
            Ftp::Macro( Ftp::Macro::Type::Clear ),
            Ftp::Macro( Ftp::Macro::Type::ShowDirectory, { destination } )
        }));

        const auto clipboardList( localClipboard.get() );
        std::transform( clipboardList.begin(), clipboardList.end(), std::back_inserter( macros ),
            []( const FileInfo& fileInfo )
            { return Ftp::Macro( Ftp::Macro::Type::Select, { fileInfo } ); } );
    }

    // decide if source has to be updated at the end of the macro
    const FileInfo sourcePath( selectionFrame( selection.front().type() ).path() );
    auto iter = std::find_if( selection.begin(), selection.end(), BaseFileInfo::SameFileFTor( sourcePath.file() ) );
    if( iter != selection.end() )
    {
        macros.append( Base::makeT<Ftp::Macro::List>(
        {
            Ftp::Macro( Ftp::Macro::Type::Clear ),
            Ftp::Macro( Ftp::Macro::Type::ShowDirectory, { sourcePath } )
        }) );
    }

    // emit macros
    if( !macros.empty() ) emit macrosRequested( macros );

    // return local clipboard, for replacement
    return localClipboard;
}

//_______________________________________________________________________
FileInfo::List MainWindow::_destinationList( const FileInfo& destination, const FileInfo::List& selection ) const
{

    FileInfo::List destinationList;
    std::transform( selection.begin(), selection.end(), std::back_inserter( destinationList ),
        [&destination]( const FileInfo& fileInfo )
        {
            // generate destination file
            FileInfo destinationFile( fileInfo );
            destinationFile.setFile( fileInfo.file().localName().addPath( destination.file() ) );
            if( destination.isLocal() ) destinationFile.setLocal();
            else destinationFile.setRemote();
            return destinationFile;
        } );

    return destinationList;

}

//___________________________________________________________________________
void MainWindow::_updatePath( const FileInfo& fileInfo )
{

    Debug::Throw() << "MainWindow::_updatePath - file: " << fileInfo << endl;
    Q_ASSERT( fileInfo.isFolder() );

    // update path
    if( fileInfo.isLocal() || ( fileInfo.isRemote() && interface_->ftp().state() == QFtp::LoggedIn ) )
    {

        emit macrosRequested(  Base::makeT<Ftp::Macro::List>({
            Ftp::Macro( Ftp::Macro::Type::Clear ),
            Ftp::Macro( Ftp::Macro::Type::ShowDirectory, { fileInfo } )
        }) );

    }

    // also update fileSystemWatcher
    if( fileInfo.isLocal() )
    {
        const auto directories( fileSystemWatcher_.directories() );
        if( !directories.isEmpty() ) fileSystemWatcher_.removePaths( directories );
        if( !fileInfo.file().isEmpty() ) fileSystemWatcher_.addPath( fileInfo.file() );
    }

}

//______________________________________________________
void MainWindow::_updateWindowTitle()
{

    File path( activeFrame().path().file() );
    path.removeTrailingSlash();
    setWindowTitle( path.localName() );

}

//___________________________________________________________________________
void MainWindow::_disableRemotePlaces()
{
    Debug::Throw( "MainWindow::_disableRemotePlaces.\n" );

    const auto fileInfoList( placesWidget_->items() );
    if( std::accumulate( fileInfoList.begin(), fileInfoList.end(), false,
        [this]( bool value, const BaseFileInfo& fileInfo )
        { return fileInfo.isRemote() && placesWidget_->setItemIsValid( fileInfo, false ) ? true:value; } ) )
    { placesWidget_->update(); }

}

//___________________________________________________________________________
void MainWindow::_requestPlacesUpdate( int type )
{

    Debug::Throw( "MainWindow::_requestPlacesUpdate.\n" );

    // loop over places widget local files and check existence
    auto macros =  Base::makeT<Ftp::Macro::List>({Ftp::Macro( Ftp::Macro::Type::Clear)});

    for( const auto& fileInfo:placesWidget_->items() )
    {

        // check file location against type
        if( type&FileInfo::Remote ) { if( fileInfo.isLocal() ) continue; }
        else if( fileInfo.isRemote() ) continue;

        macros.append( Ftp::Macro( Ftp::Macro::Type::Add, { FileInfo( fileInfo ) } ) );
    }

    macros.append( Ftp::Macro( Ftp::Macro::Type::CheckFilesExist ) );
    emit macrosRequested( macros );

}

//______________________________________________________
void MainWindow::_updateVisibilityActions()
{

    Debug::Throw( "MainWindow::_updateVisibilityActions.\n" );
    localSelectionFrame_->visibilityAction().setEnabled( remoteSelectionFrame_->visibilityAction().isChecked() );
    remoteSelectionFrame_->visibilityAction().setEnabled( localSelectionFrame_->visibilityAction().isChecked() );

    // handle focus
    if( !localSelectionFrame_->visibilityAction().isChecked() ) remoteSelectionFrame_->setFocus();
    if( !remoteSelectionFrame_->visibilityAction().isChecked() ) localSelectionFrame_->setFocus();

}

//______________________________________________________
void MainWindow::_updateNavigationActions()
{
    Debug::Throw( "MainWindow::_updateNavigationActions.\n" );
    reloadAction_->setEnabled( activeFrame().reloadAction().isEnabled() );

    previousDirectoryButton_->setEnabled( activeFrame().previousDirectoryAction().isEnabled() );
    previousDirectoryButton_->setMenu( &activeFrame().pathEditor().previousPathMenu() );

    nextDirectoryButton_->setEnabled( activeFrame().nextDirectoryAction().isEnabled() );
    nextDirectoryButton_->setMenu( &activeFrame().pathEditor().nextPathMenu() );

    parentDirectoryAction_->setEnabled( activeFrame().parentDirectoryAction().isEnabled() );
    homeDirectoryAction_->setEnabled( activeFrame().homeDirectoryAction().isEnabled() );
    return;
}

//______________________________________________________
void MainWindow::_updateViewModeActions()
{
    Debug::Throw( "MainWindow::_updateViewModeActions.\n" );
    iconsViewAction_->setChecked( activeFrame().viewMode() == FileSelectionFrame::ViewMode::Icons );
    detailsViewAction_->setChecked( activeFrame().viewMode() == FileSelectionFrame::ViewMode::Details );
}

//____________________________________________________________________________________________
void MainWindow::_updateActiveFrame( QWidget*, QWidget* widget )
{

    Debug::Throw() << "MainWindow::_updateActiveFrame -"
        << " widget: " << widget
        << " (" << (widget ? widget->metaObject()->className():"none") << ")"
        << endl;

    if( localSelectionFrame_->owns( widget ) )
    {

        localSelectionFrame_->setActive( true );
        remoteSelectionFrame_->setActive( false );
        _updateNavigationActions();
        _updateViewModeActions();
        _updateWindowTitle();
        _updateStatusBar();

    } else if(  remoteSelectionFrame_->owns( widget ) ) {

        localSelectionFrame_->setActive( false );
        remoteSelectionFrame_->setActive( true );
        _updateNavigationActions();
        _updateViewModeActions();
        _updateWindowTitle();
        _updateStatusBar();

    }

}

//____________________________________________________________________________________________
void MainWindow::_updateViewMode( FileSelectionFrame::ViewMode mode )
{
    FileSelectionFrame& selectionFrame( activeFrame() );
    selectionFrame.setViewMode( mode );
    selectionFrame.setShowNavigator( showNavigator_ && mode == FileSelectionFrame::ViewMode::Details );
    _updateViewModeActions();
}

//____________________________________________________________________________________________
void MainWindow::_updateStatusBarDelayed()
{ statusbarUpdateTimer_.start( 200, this ); }

//____________________________________________________________________________________________
void MainWindow::_updateStatusBar()
{ statusbar_->label().setTextAndUpdate( activeFrame().description() ); }

//____________________________________________________________________________________________
void MainWindow::_toggleIconsViewMode( bool value )
{

    // take action only on check-in
    if( !value ) return;

    Debug::Throw( "MainWindow::_toggleIconsViewMode.\n" );
    FileSelectionFrame& selectionFrame( activeFrame() );
    selectionFrame.setViewMode( FileSelectionFrame::ViewMode::Icons );
    selectionFrame.setShowNavigator( false );

}

//____________________________________________________________________________________________
void MainWindow::_toggleDetailsViewMode( bool value )
{
    // take action only on check-in
    if( !value ) return;

    Debug::Throw( "MainWindow::_toggleDetailsViewMode.\n" );
    FileSelectionFrame& selectionFrame( activeFrame() );
    selectionFrame.setViewMode( FileSelectionFrame::ViewMode::Details );
    selectionFrame.setShowNavigator( showNavigator_ );
}

//________________________________________________________________________
void MainWindow::_toggleShowHiddenFiles( bool value )
{
    Debug::Throw() << "MainWindow::_toggleShowHiddenFiles - " << value << endl;
    interface_->setShowHiddenFiles( value );
    localSelectionFrame_->setShowHiddenFiles( value );
    remoteSelectionFrame_->setShowHiddenFiles( value );
    XmlOptions::get().set<bool>( "SHOW_HIDDEN_FILES", value );
}

//____________________________________________________________________________________________
void MainWindow::_toggleConnection()
{

    if( !enabled_ ) return;

    if( interface_->ftp().state() == QFtp::LoggedIn ) emit macroRequested( Ftp::Macro( Ftp::Macro::Type::CloseConnection ) );
    else {

        ConnectionDialog dialog( this );
        dialog.setWindowTitle( tr( "Connections" ) );
        if( !dialog.centerOnParent().exec() )
        {
            enabled_ = false;
            toggleConnectionAction_->setChecked( false );
            enabled_ = true;
            return;
        }

        attributes_ = dialog.frame().attributes();

        emit macroRequested( Ftp::Macro( Ftp::Macro::Type::Connect, {
            FileInfo( attributes_.host() ),
            FileInfo( QString::number( attributes_.port() ) ),
            FileInfo( attributes_.user() ),
            FileInfo( attributes_.password() )
        } ) );

    }

}

//___________________________________________________________________________
void MainWindow::_abort()
{

    Debug::Throw( "MainWindow::_abort.\n" );

    // retrieve current local and remote path
    FileInfo localPath( localSelectionFrame_->path() );
    FileInfo remotePath( remoteSelectionFrame_->path() );

    // process Abort command
    interface_->clear();
    emit macrosRequested(  Base::makeT<Ftp::Macro::List>({
        Ftp::Macro( Ftp::Macro::Type::Abort ),
        Ftp::Macro( Ftp::Macro::Type::ShowDirectory, { localPath } ),
        Ftp::Macro( Ftp::Macro::Type::ShowDirectory, { remotePath } )
    } ));

    return;

}

//____________________________________________________________________________________________
void MainWindow::_fileSelected( const FileInfo& fileInfo )
{
    Debug::Throw() << "MainWindow::_fileSelected - file: " << fileInfo << endl;
    if( fileInfo.isFolder() ) _updatePath( fileInfo );
    else  _openFiles(  Base::makeT<FileInfo::List>( {fileInfo} ) );
}

//____________________________________________________________________________________________
void MainWindow::_placeSelected( const BaseFileInfo& fileInfo )
{
    Debug::Throw() << "MainWindow::_placeSelected - file: " << fileInfo << endl;
    if( !fileInfo.isFolder() ) return;

    // update path
    _updatePath( FileInfo( fileInfo ) );

    // set focus to correct selection frame
    selectionFrame( fileInfo.type() ).setFocus();

    return;
}

//_______________________________________________________________________
void MainWindow::_reload()
{ if( activeFrame().reloadAction().isEnabled() ) activeFrame().reloadAction().trigger(); }

//_______________________________________________________________________
void MainWindow::_previousDirectory()
{ if( activeFrame().previousDirectoryAction().isEnabled() ) activeFrame().previousDirectoryAction().trigger(); }

//_______________________________________________________________________
void MainWindow::_nextDirectory()
{ if( activeFrame().nextDirectoryAction().isEnabled() ) activeFrame().nextDirectoryAction().trigger(); }

//_______________________________________________________________________
void MainWindow::_parentDirectory()
{ if( activeFrame().parentDirectoryAction().isEnabled() ) activeFrame().parentDirectoryAction().trigger(); }

//_______________________________________________________________________
void MainWindow::_homeDirectory()
{ if( activeFrame().homeDirectoryAction().isEnabled() ) activeFrame().homeDirectoryAction().trigger(); }

//______________________________________________________________
void MainWindow::_updateClipboard( const FileSelectionClipboard& clipboard )
{

    Debug::Throw( "MainWindow::_updateClipboard.\n" );
    clipboard_ = clipboard;

    {
        // local frame update
        const bool enabled( !clipboard_.isEmpty() );
        localSelectionFrame_->pasteAction().setEnabled( enabled );
        localSelectionFrame_->pasteInFolderAction().setEnabled( enabled );

        if( enabled && !clipboard_.isEmpty() ) localSelectionFrame_->pasteAction().setText( tr( "Paste %1" ).arg( clipboard_.description() ) );
        else localSelectionFrame_->pasteAction().setText( tr( "Paste" ) );
        localSelectionFrame_->model().setClipboard( clipboard );
    }

    {
        // remote frame update
        const bool connected( interface_->ftp().state() == QFtp::LoggedIn );
        const bool enabled( connected && !clipboard_.isEmpty() && (clipboard_.get().front().isLocal() || clipboard_.mode() == FileSelectionClipboard::Cut ) );
        remoteSelectionFrame_->pasteAction().setEnabled( enabled );
        remoteSelectionFrame_->pasteInFolderAction().setEnabled( enabled );

        if( enabled && !clipboard_.isEmpty() ) remoteSelectionFrame_->pasteAction().setText( tr( "Paste %1" ).arg( clipboard_.description() ) );
        else remoteSelectionFrame_->pasteAction().setText( tr( "Paste" ) );
        remoteSelectionFrame_->model().setClipboard( clipboard );
    }

    // also update QClipboard
    if( !clipboard_.isEmpty() )
    {

        QStringList files;
        const auto fileInfoList( clipboard_.get() );
        std::transform( fileInfoList.begin(), fileInfoList.end(), std::back_inserter( files ),
            []( const FileInfo& fileInfo ) { return (fileInfo.isRemote() ? Ftp::Defines::remotePrefix : Ftp::Defines::localPrefix) + "//" + fileInfo.file(); } );

        qApp->clipboard()->setText( files.join( " " ) );

    }

}

//_______________________________________________________________________
void MainWindow::_pasteSelection( const FileInfo& constDestination, const FileInfo::List& selection )
{

    Debug::Throw( "MainWindow::_pasteSelection.\n" );
    if( selection.empty() ) return;

    // make sure valid destination is set
    FileInfo destination( constDestination );
    if( destination.file().isEmpty() ) destination = selectionFrame( destination.location() ).path();

    // check transfer
    if( !_checkTransfer( destination, selection ) ) return;

    // disable remote-to-remote copy (not supported by qft)
    menuActions_[0]->setEnabled( !( destination.isRemote() && selection.front().isRemote() ) );

    // open menu
    QMenu menu( this );
    menu.addActions( menuActions_ );
    menu.ensurePolished();
    QAction* action( menu.exec( QCursor::pos() ) );
    if( action == menuActions_[0] ) _paste( destination, selection );
    else if( action == menuActions_[1] ) _move( destination, selection );
    else return;

}

//_______________________________________________________________________
void MainWindow::_pasteClipboard( const FileInfo& destination )
{

    Debug::Throw( "MainWindow::_pasteClipboard.\n" );
    if( clipboard_.isEmpty() ) return;
    if( !_checkTransfer( destination, clipboard_.get() ) ) return;

    // perform command
    if( clipboard_.mode() == FileSelectionClipboard::Copy ) _paste( destination, clipboard_.get() );
    else clipboard_  = _move( destination, clipboard_.get() );

}

//______________________________________________________________
void MainWindow::_openFiles( const FileInfo::List& fileInfoList )
{
    Debug::Throw( "MainWindow::_openFiles.\n" );

    OpenWithDialog dialog( this );
    dialog.setWindowTitle( tr( "Open Files" ) );

    {
        // add files
        File::List files;
        std::transform( fileInfoList.begin(), fileInfoList.end(), std::back_inserter( files ),
            []( const FileInfo& fileInfo ) { return fileInfo.file(); } );
        dialog.setFiles( files );
    }

    dialog.setOptionName( "APPLICATIONS" );
    dialog.setAutoOpen( false );
    dialog.realizeWidget();

    if( !dialog.exec() ) return;

    // check
    if( !dialog.isCommandValid() || (dialog.command().isEmpty() && !dialog.isCommandDefault() ) )
    {
        InformationDialog( this, tr( "No command specified to open the selected files. <Open> canceled." ) ).exec();
        return;
    }

    // retrieve local and remote path
    FileInfo localPath( localSelectionFrame_->path() );
    FileInfo remotePath( remoteSelectionFrame_->path() );

    // create list of macros
    auto macros =  Base::makeT<Ftp::Macro::List>({ Ftp::Macro( Ftp::Macro::Type::Clear ) });

    // loop over selected items and add to interface
    std::transform( fileInfoList.begin(), fileInfoList.end(), std::back_inserter( macros ),
        []( const FileInfo& fileInfo )
        { return Ftp::Macro( Ftp::Macro::Type::Add, { fileInfo } ); } );

    // submit open command
    if( dialog.isCommandDefault() ) macros.append( Ftp::Macro( Ftp::Macro::Type::Open, { localPath, remotePath } ) );
    else {

        FileInfo command( dialog.command() );
        macros.append( Ftp::Macro( Ftp::Macro::Type::Open, { localPath, remotePath, command } ) );

    }

    // emit
    emit macrosRequested( macros );

}

//____________________________________________________________________________________________
void MainWindow::_macroStarted( const Ftp::Macro& macro )
{

    Debug::Throw( "MainWindow::_macroStarted.\n" );
    switch( macro.type() )
    {
        case Ftp::Macro::Type::Remove:
        case Ftp::Macro::Type::Open:
        case Ftp::Macro::Type::CopyRecursive:
        statusbar_->showProgressBar();
        taskBarProgressNotifications_->setValue(0);
        taskBarProgressNotifications_->setProgressVisible( true );
        break;

        default: break;
    }

}

//____________________________________________________________________________________________
void MainWindow::_macroCompleted( const Ftp::Macro& macro, bool )
{
    Debug::Throw( "MainWindow::_macroCompleted.\n" );

    switch( macro.type() )
    {

        case Ftp::Macro::Type::Remove:
        case Ftp::Macro::Type::Open:
        statusbar_->hideProgressBar();
        break;

        case Ftp::Macro::Type::CopyRecursive:
        {

            statusbar_->hideProgressBar();
            taskBarProgressNotifications_->setValue(0);
            taskBarProgressNotifications_->setProgressVisible( false );

            // adjust column sizes
            if( TreeView* treeView = qobject_cast<TreeView*>( &selectionFrame( FileInfo( macro.argument(0) ).type() ).list() ) )
            { treeView->resizeColumnToContents( FileInfoModel::FileName ); }
            break;

        }

        case Ftp::Macro::Type::CheckConflicts:
        {

            // retrieve files and check if there is a conflict
            FileInfo::List& files( interface_->filesRecursive() );
            if( !std::count_if( files.begin(), files.end(), FileInfo::ConflictFTor() ) ) break;


            // create temporary list
            FileInfo::List local;
            for( const auto& fileInfo:files )
            { if( fileInfo.isDocument() && fileInfo.isConflict() ) local.append( fileInfo ); }

            // create dialog
            ConflictDialog dialog( this );
            dialog.setWindowTitle( tr( "Conflicts" ) );
            dialog.setFiles(
                macro.argument(0).file(),
                macro.argument(1).file(),
                local );

            if( dialog.centerOnParent().exec() == QDialog::Rejected ) files.clear();
            else dialog.updateFiles( files );
            break;

        }

        case Ftp::Macro::Type::CheckFilesCopy:
        case Ftp::Macro::Type::CheckFilesMoved:
        {

            // retrieve files and check if copy has succedded
            const FileInfo::List& files( interface_->filesRecursive() );

            FileInfo::List local;
            for( const auto& fileInfo:files )
            { if( fileInfo.isFailed() ) local.append( fileInfo ); }

            // check if there are some failed files
            if( local.empty() ) break;

            // create dialog
            const File sourcePath( macro.argument(0).file() );
            const File destinationPath( macro.argument(1).file() );

            CheckFilesDialog dialog( this );
            dialog.setWindowTitle( tr( "Error" ) );
            dialog.setFiles(
                sourcePath,
                destinationPath,
                local );

            // this needs double check. Probably broken
            if( dialog.centerOnParent().exec() == QDialog::Accepted )
            {

                // get selection
                if( !(local = dialog.selectedFiles()).empty() )
                {
                    if( macro.type() == Ftp::Macro::Type::CheckFilesCopy ) _paste( FileInfo( destinationPath ), local );
                    else _move( FileInfo( destinationPath ), local );
                }

            }
            break;

        }

        case Ftp::Macro::Type::CheckFilesExist:
        {

            // retrieve files and check if copy has succedded
            const auto files( interface_->filesRecursive() );
            if( std::accumulate( files.begin(), files.end(), false,
                [this]( bool value, const FileInfo& fileInfo )
                { return placesWidget_->setItemIsValid( fileInfo, fileInfo.isSuccess() ) ? true:value; } ) )
            { placesWidget_->update(); }

            break;
        }


        case Ftp::Macro::Type::ClearSelection:
        selectionFrame( macro.argument(0).type() ).clearSelection();
        break;

        case Ftp::Macro::Type::Select:
        {
            const FileInfo fileInfo( macro.argument(0) );
            selectionFrame( fileInfo.type() ).select( fileInfo );
            break;
        }


        default:
        break;

    }

    return;

}

//____________________________________________________________________________________________
void MainWindow::_commandCompleted( const Ftp::Command& command, bool )
{

    Debug::Throw( "MainWindow::_commandCompleted.\n" );
    switch( command.type() )
    {

        case Ftp::Command::Type::Copy:
        {
            // retrieve fileInfo from command
            // and update the corresponding list
            FileInfo fileInfo( command.argument(1) );
            selectionFrame( fileInfo.type() ).add( fileInfo );
            break;
        }

        case Ftp::Command::Type::Remove:
        {
            FileInfo fileInfo( command.argument(0) );
            if( fileInfo.isLocal() ) localSelectionFrame_->remove( fileInfo );
            else {
                remoteSelectionFrame_->remove( fileInfo );
                if( fileInfo.isFolder() && placesWidget_->setItemIsValid( fileInfo, false ) )
                { placesWidget_->update(); }

            }

            break;

        }

        case Ftp::Command::Type::CreateDirectory:
        {

            FileInfo fileInfo( command.argument(0) );
            if( fileInfo.isLocal() ) localSelectionFrame_->add( fileInfo );
            else {

                // add to remote frame and update places item if needed
                remoteSelectionFrame_->add( fileInfo );
                if( placesWidget_->setItemIsValid( fileInfo, true ) )
                { placesWidget_->update(); }

            }
            break;

        }


        default:
        break;
    }
}

//___________________________________________________________________________
void MainWindow::_updateContent( const FileInfo& path, const FileInfo::List& files )
{

    Debug::Throw( "MainWindow::_updateContent.\n" );
    FileSelectionFrame& selectionFrame( this->selectionFrame( path.type() ) );

    // build file info list with local filenames
    FileInfo::List local;
    for( const auto& fileInfo:files )
    { if( fileInfo.isLocal() == path.isLocal() ) local.append( fileInfo ); }

    // clear the model if path differs from selectionFrame path
    if( path.file() != selectionFrame.path().file() )
    {

        selectionFrame.setPath( path.file() );
        selectionFrame.model().clear();
        selectionFrame.model().set( local );
        _updateWindowTitle();

    } else {

        selectionFrame.model().update( local );

    }

    // resize columns
    if( TreeView* treeView = qobject_cast<TreeView*>( &selectionFrame.list() ) )
    { treeView->resizeColumnToContents( FileInfoModel::FileName ); }

    // navigator
    selectionFrame.setShowNavigator( showNavigator_ && selectionFrame.viewMode() == FileSelectionFrame::ViewMode::Details );

}

//___________________________________________________________________________
void MainWindow::_updateFileInfo( const FileInfo& info )
{
    Debug::Throw() << "MainWindow::_updateFileInfo - file: " << info.file() << (info.isRemote() ? " (remote)":" (local)") << endl;
    selectionFrame( info.type() ).add( info );
}

//___________________________________________________________________________
void MainWindow::_requestPlacesUpdate( const BaseFileInfo& fileInfo )
{

    Debug::Throw( "MainWindow::_requestPlacesUpdate.\n" );

    // loop over places widget local files and check existence
    emit macrosRequested( Base::makeT<Ftp::Macro::List>(
    {
        Ftp::Macro( Ftp::Macro::Type::Clear ),
        Ftp::Macro( Ftp::Macro::Type::Add, { FileInfo( fileInfo ) } ),
        Ftp::Macro( Ftp::Macro::Type::CheckFilesExist )
    }) );

}

//___________________________________________________________________________
void MainWindow::_stateChanged( int state )
{

    Debug::Throw( "MainWindow::_stateChanged.\n" );

    enabled_ = false;
    QString buffer;
    QTextStream what( &buffer );
    switch( state )
    {

        case QFtp::Unconnected:
        {
            remoteSelectionFrame_->clear();
            enabled_ = false;
            toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Disconnect ) );
            toggleConnectionAction_->setText( tr( "Connect" ) );
            enabled_ = true;
            _updateButtons();
            remoteSelectionFrame_->setEnabled( false );
            buffer = tr( "Not connected" );

            // disable remote places
            _disableRemotePlaces();
            break;

        }

        case QFtp::HostLookup:
        {
            enabled_ = false;
            toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Disconnect ) );
            toggleConnectionAction_->setText( tr( "Connect" ) );
            enabled_ = true;

            buffer = tr( "Looking for host '%1'" ).arg( attributes_.host() );
            break;
        }

        case QFtp::Connecting:
        {
            enabled_ = false;
            toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Disconnect ) );
            toggleConnectionAction_->setText( tr( "Connect" ) );
            enabled_ = true;

            buffer = tr( "Connecting" );
            break;
        }

        case QFtp::Connected:
        {
            enabled_ = false;
            toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Disconnect ) );
            toggleConnectionAction_->setText( tr( "Connect" ) );
            enabled_ = true;

            buffer = tr( "Connected" );
            break;
        }

        case QFtp::LoggedIn:
        {
            remoteSelectionFrame_->setPath( File("/") );
            _fileSelected( FileInfo( "/", FileInfo::Folder|FileInfo::Remote ) );
            _updateButtons();
            remoteSelectionFrame_->setEnabled( true );
            toggleConnectionAction_->setIcon( IconEngine::get( IconNames::Connect ) );
            toggleConnectionAction_->setText( tr( "Disconnect" ) );

            buffer = tr( "Logged in as '%1'" ).arg( attributes_.user() );

            // disable remote places
            _requestRemotePlacesUpdate();
            break;

        }

        default:
        break;

    }

    enabled_ = true;
    if( buffer.isEmpty() ) return;

    //* retrieve StateFrame
    statusbar_->label(0).setText( buffer, true );

}

//___________________________________________________________________________
void MainWindow::_updateButtons()
{

    Debug::Throw( "MainWindow::_updateButtons.\n" );
    const bool connected( interface_->ftp().state() == QFtp::LoggedIn );
    remoteSelectionFrame_->homeDirectoryAction().setEnabled( connected );
    remoteSelectionFrame_->reloadAction().setEnabled( connected );
    remoteSelectionFrame_->newFolderAction().setEnabled( connected );

    if( !connected )
    {
        remoteSelectionFrame_->previousDirectoryAction().setEnabled( false );
        remoteSelectionFrame_->nextDirectoryAction().setEnabled( false );
        remoteSelectionFrame_->parentDirectoryAction().setEnabled( false );
        remoteSelectionFrame_->renameAction().setEnabled( false );
        remoteSelectionFrame_->removeAction().setEnabled( false );
        remoteSelectionFrame_->openAction().setEnabled( false );
    }

    _updateClipboard( clipboard_ );

}

//___________________________________________________________________________
void MainWindow::_updateBusyButtons( bool state )
{

    Debug::Throw( "MainWindow::_updateBusyButtons.\n" );
    abortAction_->setEnabled( state );

}

//__________________________________________________________________________
void MainWindow::_showCommandDialog()
{

    Debug::Throw( "MainWindow::_showCommandDialog.\n" );
    XmlOptions::get().set<bool>( "SHOW_COMMAND_LIST", true );
    commandDialog_->show();
    commandDialog_->raise();

}

//_______________________________________________
void MainWindow::_commandDialogHidden()
{
    Debug::Throw( "Application::_commandDialogHidden.\n" );
    XmlOptions::get().set<bool>( "SHOW_COMMAND_LIST", false );
}

//__________________________________________________________________________
void MainWindow::_showLogDialog()
{

    Debug::Throw( "MainWindow::_showLogDialog.\n" );
    XmlOptions::get().set<bool>( "SHOW_LOG", true );
    logDialog().show();
    logDialog().raise();

}

//_______________________________________________
void MainWindow::_logDialogHidden()
{
    Debug::Throw( "MainWindow::_logDialogHidden.\n" );
    XmlOptions::get().set<bool>( "SHOW_LOG", false );
}

//____________________________________________________________________________________________
void MainWindow::_toggleShowNavigator( bool value )
{

    Debug::Throw() << "MainWindow::_toggleShowNavigator - " << value << endl;
    if( showNavigator_ == value ) return;
    showNavigator_ = value;
    for( const auto& selectionFrame:selectionFrames() )
    { selectionFrame->setShowNavigator( showNavigator_ && selectionFrame->viewMode() == FileSelectionFrame::ViewMode::Details ); }

}

//_______________________________________________
void MainWindow::_fileInformation( const FileInfo& fileInfo, const QIcon& icon )
{

    FileInformationDialog dialog( this, fileInfo );
    dialog.setIcon( icon );

    connect( &dialog, SIGNAL(macrosRequested(Ftp::Macro::List)), interface_, SLOT(addMacros(Ftp::Macro::List)) );
    connect( interface_, SIGNAL(sizeAvailable(FileInfo,qint64)), &dialog, SLOT(updateSize(FileInfo,qint64)) );
    connect( interface_, SIGNAL(macroStarted(Ftp::Macro)), &dialog, SLOT(macroStarted(Ftp::Macro)) );
    connect( interface_, SIGNAL(macroCompleted(Ftp::Macro,bool)), &dialog, SLOT(macroCompleted(Ftp::Macro)) );

    dialog.exec();

}

//____________________________________________________________________________________________
void MainWindow::_find( TextSelection selection )
{ activeFrame().find( selection ); }


//____________________________________________________________________________________________
void MainWindow::_restoreFocus()
{ activeFrame().list().setFocus(); }

//____________________________________________________________________________________________
void MainWindow::_updateConfiguration()
{
    Debug::Throw( "MainWindow::_updateConfiguration.\n" );

    // command dialog
    if( XmlOptions::get().get<bool>( "SHOW_COMMAND_LIST" ) ) _showCommandDialog();
    else commandDialog_->hide();

    // log dialog
    if( XmlOptions::get().get<bool>( "SHOW_LOG" ) ) _showLogDialog();
    else logDialog().hide();

    // navigator
    _toggleShowNavigator( XmlOptions::get().get<bool>( "SHOW_NAVIGATOR" ) );

    // hidden files
    showHiddenFilesAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_HIDDEN_FILES" ) );

    // clear cached icons
    iconProvider_->clear();

}

//_______________________________________________________________________
void MainWindow::_installActions()
{
    Debug::Throw( "MainWindow::_installActions.\n" );

    // connect
    addAction( toggleConnectionAction_ = new QAction( IconEngine::get( IconNames::Disconnect ), tr( "Connect" ), this ) );
    connect( toggleConnectionAction_, SIGNAL(triggered()), SLOT(_toggleConnection()) );

    // home directory
    addAction( homeDirectoryAction_ = new QAction( IconEngine::get( IconNames::Home ), tr( "Home" ), this ) );
    connect( homeDirectoryAction_, SIGNAL(triggered()), SLOT(_homeDirectory()) );
    homeDirectoryAction_->setToolTip( tr( "Change path to user home" ) );

    // previous directory
    previousDirectoryButton_ = new CustomToolButton( this );
    previousDirectoryButton_->setIcon( IconEngine::get( IconNames::PreviousDirectory ) );
    previousDirectoryButton_->setText( tr( "Back" ) );
    previousDirectoryButton_->setToolTip( tr( "Change path to previous directory (from history)" ) );
    connect( previousDirectoryButton_, SIGNAL(clicked()), SLOT(_previousDirectory()) );

    // next directory
    nextDirectoryButton_ = new CustomToolButton( this );
    nextDirectoryButton_->setIcon( IconEngine::get( IconNames::NextDirectory ) );
    nextDirectoryButton_->setText( tr( "Forward" ) );
    nextDirectoryButton_->setToolTip( tr( "Change path to next directory (from history)" ) );
    connect( nextDirectoryButton_, SIGNAL(clicked()), SLOT(_nextDirectory()) );

    // parent directory in tree
    addAction( parentDirectoryAction_ = new QAction( IconEngine::get( IconNames::Parent ), tr( "Parent Directory" ), this ) );
    parentDirectoryAction_->setIconText( tr( "Up" ) );
    connect( parentDirectoryAction_, SIGNAL(triggered()), SLOT(_parentDirectory()) );
    parentDirectoryAction_->setToolTip( tr( "Change path to parent directory" ) );

    // reload
    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(_reload()) );
    reloadAction_->setShortcut( QKeySequence::Refresh );
    reloadAction_->setToolTip( tr( "Reload" ) );

    // abort
    addAction( abortAction_ = new QAction( IconEngine::get( IconNames::Abort ), tr( "Abort" ), this ) );
    abortAction_->setToolTip( tr( "Abort transfer" ) );
    connect( abortAction_, SIGNAL(triggered()), SLOT(_abort()) );

    // show command
    addAction( showCommandDialogAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Show Command List" ), this ) );
    connect( showCommandDialogAction_, SIGNAL(triggered()), SLOT(_showCommandDialog()) );

    // show log
    addAction( showLogDialogAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Show Log" ), this ) );
    connect( showLogDialogAction_, SIGNAL(triggered()), SLOT(_showLogDialog()) );

    // show hidden files
    addAction( showHiddenFilesAction_ = new QAction( tr( "Show Hidden Files" ), this ) );
    showHiddenFilesAction_->setCheckable( true );
    showHiddenFilesAction_->setChecked( false );

    // icon/list view actions
    iconsViewAction_ = new QAction( IconEngine::get( IconNames::IconsView ), tr( "Icons view mode" ), this );
    iconsViewAction_->setCheckable( true );
    connect( iconsViewAction_, SIGNAL(toggled(bool)), SLOT(_toggleIconsViewMode(bool)) );

    detailsViewAction_ = new QAction( IconEngine::get( IconNames::DetailsView ), tr( "Details view mode" ), this );
    detailsViewAction_->setCheckable( true );
    connect( detailsViewAction_, SIGNAL(toggled(bool)), SLOT(_toggleDetailsViewMode(bool)) );

    QActionGroup* viewModeActionGroup = new QActionGroup( this );
    viewModeActionGroup->setExclusive( true );
    viewModeActionGroup->addAction( iconsViewAction_ );
    viewModeActionGroup->addAction( detailsViewAction_ );

    // menu actions
    QAction* separator;
    QAction* action;
    menuActions_ = Base::makeT<ActionList>({
        new QAction( IconEngine::get( IconNames::Copy ), tr( "Copy Here" ), this ),
        new QAction( IconEngine::get( IconNames::Move ), tr( "Move Here" ), this ),
        separator = new QAction( this ),
        action = new QAction( IconEngine::get( IconNames::DialogCancel ), tr( "Cancel" ), this )
    });

    // adjust
    separator->setSeparator( true );
    action->setShortcut( Qt::Key_Escape );

}

//______________________________________________________________________
void MainWindow::_createFindWidget()
{

    Debug::Throw( "MainWindow::_createFindWidget.\n" );
    if( !findWidget_ )
    {

        findWidget_ = new BaseFindWidget( rightContainer_ );
        rightContainer_->layout()->addWidget( findWidget_ );
        connect( findWidget_, SIGNAL(find(TextSelection)), SLOT(_find(TextSelection)) );
        connect( this, SIGNAL(matchFound()), findWidget_, SLOT(matchFound()) );
        connect( this, SIGNAL(noMatchFound()), findWidget_, SLOT(noMatchFound()) );
        connect( &findWidget_->closeButton(), SIGNAL(clicked()), SLOT(_restoreFocus()) );
        findWidget_->hide();

    }

    return;

}
